using System;
using System.Collections;
using MacGruber;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class CamAngle
    {
        private Pose pose;
        public int id;
        public byte[] img;
        private Vector3 camPosition;
        private float rigRotation;
        private float camRotation;
        private float focus;
        public bool hasCam = true;
        
        public EventTrigger onCamEnter;
        public EventTrigger onCamExit;

        private static SuperController sc = SuperController.singleton;
        private static WaitForFixedUpdate wait = new WaitForFixedUpdate();

        public CamAngle(Pose pose, byte[] img)
        {
            this.pose = pose;
            this.img = img;
            id = pose.camAngles.Count;
            GetAngle();
            onCamEnter = new EventTrigger(PoseMe.singleton, "On Angle Enter");
            onCamExit = new EventTrigger(PoseMe.singleton, "On Angle Exit");
            PoseMe.onAngleEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onAngleExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onAngleEnterTriggerButton.button.onClick.AddListener(onCamEnter.OpenPanel);
            PoseMe.onAngleExitTriggerButton.button.onClick.AddListener(onCamExit.OpenPanel);
        }
        
        public CamAngle(Pose pose, JSONClass jc)
        {
            this.pose = pose;
            id = pose.camAngles.Count;
            onCamEnter = new EventTrigger(PoseMe.singleton, "On Angle Enter");
            onCamExit = new EventTrigger(PoseMe.singleton, "On Angle Exit");
            PoseMe.onAngleEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onAngleExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onAngleEnterTriggerButton.button.onClick.AddListener(onCamEnter.OpenPanel);
            PoseMe.onAngleExitTriggerButton.button.onClick.AddListener(onCamExit.OpenPanel);
            onCamEnter.RestoreFromJSON(jc, PoseMe.singleton.subScenePrefix, false, true);
            onCamExit.RestoreFromJSON(jc, PoseMe.singleton.subScenePrefix, false, true);
            Load(jc);
        }

        private void GetAngle()
        {
            camPosition = pose.storedWorldToLocalMatrix.MultiplyPoint(GetCamPosition());
            rigRotation = (pose.storedWorldToLocalMatrix.rotation * PoseMe.cameraRig.rotation).eulerAngles.y;
            camRotation = PoseMe.camera.localEulerAngles.x;
            focus = SuperController.singleton.focusDistance;
            hasCam = true;
        }

        public void Update(byte[] img)
        {
            GetAngle();
            this.img = img;
        }

        public void Apply()
        {
            // PoseMe.camera.position.y.Print();
            if(!hasCam) return;
            Pose.smoothCam.Stop();
            var refMatrix = pose.refMatrix;
            var rY = Quaternion.Euler(new Vector3(0f, refMatrix.rotation.eulerAngles.y, 0f));
            var mtx = Matrix4x4.TRS(refMatrix.GetColumn(3), rY, Vector3.one);
            
            var camPos = mtx.MultiplyPoint(camPosition);
            
            var p0 = GetCamPosition();
            var p2 = camPos;
            var delta = (p2 - p0).normalized;
            var dot = Vector3.Dot(delta, PoseMe.camera.forward);
            var dot3 = dot * dot * dot;
            var p1 = .5f * (p0 + p2) + dot3*PoseMe.smoothCamBezierStrength.val*Vector3.Cross(delta, Vector3.Cross(delta, PoseMe.camera.forward)).normalized;
            bezier.p0 = p0;
            bezier.p1 = p1;
            bezier.p2 = p2;
            // (dot3*dot3*dot3).Print();
            // PoseMe.test1.transform.position = p0;
            // PoseMe.test2.transform.position = p1;
            // PoseMe.test3.transform.position = p2;
            
            var rigRot = refMatrix.rotation * Quaternion.Euler(0f, rigRotation, 0f);
            switch (PoseMe.camMode.val)
            {
                case "None":
                {
                    break;
                }
                case "Exponential":
                {
                    Pose.smoothCam = SmoothCamRestoreExp(camPos, rigRot.eulerAngles.y, camRotation, focus).Start();
                    break;
                }
                case "Bezier":
                {
                    Pose.smoothCam = SmoothCamRestoreBezier(rigRot.eulerAngles.y, camRotation, focus, 1f+Mathf.Min(.3f, Mathf.Abs(dot3*dot3*dot3))).Start();
                    break;
                }
                case "Linear":
                {
                    Pose.smoothCam = SmoothCamRestoreLin(camPos, rigRot.eulerAngles.y, camRotation, focus).Start();
                    break;
                }
                default:
                {
                    PoseMe.cameraRig.rotation = Quaternion.Euler(0f, rigRot.eulerAngles.y, 0f);;
                    PoseMe.camera.localEulerAngles = new Vector3(camRotation,0f,0f);
                    sc.focusDistance = focus;
                    SetCamPosition(camPos);
                    sc.SyncMonitorRigPosition();
                    break;
                }
            }
            pose.currentAngle = this;
            PoseMe.camChooser.valNoCallback = id.ToString();
            pose.SetUIImage(img);
            pose.SetButtonText();
            if(pose.currentAngle != null && pose.currentAngle != this) pose.currentAngle.onCamExit.Trigger();
            onCamEnter.Trigger();
            PoseMe.onAngleEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onAngleExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onAngleEnterTriggerButton.button.onClick.AddListener(onCamEnter.OpenPanel);
            PoseMe.onAngleExitTriggerButton.button.onClick.AddListener(onCamExit.OpenPanel);

            

        }

        public static Bezier bezier = new Bezier();
        
        private IEnumerator SmoothCamRestoreExp(Vector3 camPos, float rigRotY, float camRotX, float focus)
        {
            var sc = SuperController.singleton;
            var deltaT = PoseMe.smoothCamSpeedJ.val * Time.fixedDeltaTime;
            var rigRot = Quaternion.Euler(0f, rigRotY, 0f);
            var camRot = Quaternion.Euler(camRotX, 0f, 0f);
            var curCamPos = GetCamPosition();
            float t = 0f;
            while (!smoothCamAbort && (Mathf.Abs(Quaternion.Dot(camRot, PoseMe.camera.localRotation)) < .99999999f  ||
                                       Vector3.SqrMagnitude(curCamPos - camPos) > 1e-4f ||
                                       Mathf.Abs(Quaternion.Dot(rigRot, PoseMe.cameraRig.rotation)) < .99999999f  ||
                                       Mathf.Abs(sc.focusDistance - focus) > 1e-3f))
            {
                sc.focusDistance = Mathf.Lerp(sc.focusDistance, focus, deltaT);
                curCamPos = Vector3.Lerp(curCamPos, camPos, 1.25f * deltaT);
                SetCamPosition(curCamPos);
                PoseMe.cameraRig.rotation = Quaternion.Slerp(PoseMe.cameraRig.rotation, rigRot, deltaT);
                PoseMe.camera.localRotation = Quaternion.Slerp(PoseMe.camera.localRotation, camRot, deltaT);
                yield return wait;
            }
        }
        
        private IEnumerator SmoothCamRestoreBezier(float rigRotY, float camRotX, float focus, float rotationSpeedMult)
        {
            var speed = (PoseMe.cinematicEnabled.val && PoseMe.cinematicCamMode > 0)
                ? PoseMe.currentCamSpeed
                : PoseMe.smoothCamSpeedJ.val;
            var deltaT = speed * Time.fixedDeltaTime;
            // var deltaT2 = deltaT * rotationSpeedMult;
            var rigRot = Quaternion.Euler(0f, rigRotY, 0f);
            var camRot = Quaternion.Euler(camRotX, 0f, 0f);
            var curCamPos = GetCamPosition();
            float t = 0f;
            int steps = 5;
            float a = 0;
            // a = deltaT;
            while (!smoothCamAbort && (Mathf.Abs(Quaternion.Dot(camRot, PoseMe.camera.localRotation)) < .99999999f  ||
                                       t < .999f ||
                                       Mathf.Abs(Quaternion.Dot(rigRot, PoseMe.cameraRig.rotation)) < .99999999f  ||
                                       Mathf.Abs(sc.focusDistance - focus) > 1e-3f))
            {
                if (steps > 0)
                {
                    a = deltaT / steps;
                }
                sc.focusDistance = Mathf.Lerp(sc.focusDistance, focus, a);
                t = Mathf.Lerp(t, 1f, a);
                var p = bezier.Evaluate(t);
                SetCamPosition(p);
                PoseMe.cameraRig.rotation = Quaternion.Slerp(PoseMe.cameraRig.rotation, rigRot, a*rotationSpeedMult);
                PoseMe.camera.localRotation = Quaternion.Slerp(PoseMe.camera.localRotation, camRot, a*rotationSpeedMult);
                steps--;
                yield return wait;
            }
        }
        
        private IEnumerator SmoothCamRestoreLin(Vector3 camPos, float rigRotY, float camRotX, float focus)
        {
            var camPosCur = PoseMe.camera.position;
            var rigRotCur = PoseMe.cameraRig.rotation;
            var camRotCur = PoseMe.camera.localRotation;
            var focusCur = sc.focusDistance;
            var rigRot = Quaternion.Euler(0f, rigRotY, 0f);
            var camRot = Quaternion.Euler(camRotX, 0f, 0f);
            float t = 0f;
            while (!smoothCamAbort && t < 1f)
            {
                SuperController.singleton.focusDistance = Mathf.Lerp(focusCur, focus, t);
                t += .5f*PoseMe.smoothCamSpeedJ.val*Time.fixedDeltaTime;
                if (t > 1f) t = 1f;
                SetCamPosition(Vector3.Lerp(camPosCur, camPos, t));
                PoseMe.cameraRig.rotation = Quaternion.Slerp(rigRotCur, rigRot, t);
                PoseMe.camera.localRotation = Quaternion.Slerp(camRotCur, camRot, t);
                yield return wait;
            }
        }

        private void SetCamPosition(Vector3 val)
        {
            sc.playerHeightAdjust = val.y - 1.6f;
            val.y = 0f;
            PoseMe.cameraRig.position = val;
            sc.SyncMonitorRigPosition();
        }

        private Vector3 GetCamPosition()
        {
            return PoseMe.cameraRig.position + new Vector3(0f, PoseMe.camera.position.y, 0f);
        }
        
        private bool smoothCamAbort => !PoseMe.buttonHovered && 
                                       (Input.GetKeyDown(KeyCode.Mouse1) ||
                                       Input.GetKeyDown(KeyCode.Mouse2) ||
                                       Input.GetAxis("Mouse ScrollWheel") != 0f ||
                                       Input.GetKeyDown(KeyCode.Tab) ||
                                       Input.GetKeyDown(KeyCode.W) ||
                                       Input.GetKeyDown(KeyCode.A) ||
                                       Input.GetKeyDown(KeyCode.S) ||
                                       Input.GetKeyDown(KeyCode.D));
        
        public JSONClass Store()
        {
            var jc = new JSONClass
            {
                ["image"] = Convert.ToBase64String(img),
                ["camPosition"] = camPosition.ToJA(),
                ["rigRotation"] = {AsFloat = rigRotation},
                ["cameraRotation"] = {AsFloat = camRotation},
                ["focusDistance"] = {AsFloat = focus},
                [onCamEnter.Name] = onCamEnter.GetJSON(PoseMe.singleton.subScenePrefix),
                [onCamExit.Name] = onCamExit.GetJSON(PoseMe.singleton.subScenePrefix),
            };
            return jc;
        }

        public void Load(JSONClass jc)
        {
            if (jc.HasKey("image")) img = Convert.FromBase64String(jc["image"]);
            if (jc.HasKey("rigPosition"))
            {
                var rigPosition = jc["rigPosition"].AsArray.ToV3();
                camPosition = new Vector3(rigPosition.x, rigPosition.y + jc["playerHeight"].AsFloat + 1.6f, rigPosition.z);
            }
            else camPosition = jc["camPosition"].AsArray.ToV3();
            rigRotation = jc["rigRotation"].AsFloat;
            camRotation = jc["cameraRotation"].AsFloat;
            focus = jc["focusDistance"].AsFloat;
            onCamEnter.RestoreFromJSON(jc, PoseMe.singleton.subScenePrefix, false, true);
            onCamExit.RestoreFromJSON(jc, PoseMe.singleton.subScenePrefix, false, true);
        }
    }
}