using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace CheesyFX
{
    public static class IdleUIProvider
    {
        public static bool uiOpen;
        private static MyUIDynamicToggle forceEnabledUid;
        private static MyUIDynamicToggle torqueEnabledUid;
        private static MyUIDynamicToggle idlesEnabledUid;
        private static UIDynamicSlider forceUid;
        private static UIDynamicSlider torqueUid;
        private static UIDynamicV3Slider directionalForceUid;
        private static UIDynamicV3Slider directionalTorqueUid;
        private static UIDynamicToggleArray forceOnesidedUid;
        private static UIDynamicToggleArray torqueOnesidedUid;

        public static JSONStorableString info = new JSONStorableString("info", "");

        private static UIDynamicTabBar tabbar;
        private static LimbIdle _currentLimbIdle;
        private static LimbIdle currentLimbIdle
        {
            get { return _currentLimbIdle; }
            set
            {
                if (_currentLimbIdle == value) return;
                if (_currentLimbIdle != null) _currentLimbIdle.isEdited = false;
                _currentLimbIdle = value;
                _currentLimbIdle.isEdited = true;
            }
        }
        public static int lastTabId;
        private static bool axisShown;
        

        private static List<object> UIElements = new List<object>();
        

        public static void CreateUI()
        {
            PoseMe.singleton.ClearUI();
            UIElements.Clear();
            var button = PoseMe.singleton.CreateButton("Return");
            button.buttonColor = PoseMe.navColor;
            button.button.onClick.AddListener(
                () =>
                {
                    AxisSetActive(false);
                    PoseMe.singleton.ClearUI();
                    uiOpen = false;
                    PoseMe.singleton.CreateUI();
                    // if(PoseMe.currentPose != null) PoseMe.currentPose.SetPreviewImage();
                });
            UIElements.Add(button);
            PoseMe.poseChooser.CreateUI(rightSide:true).ForceHeight(50f);
            tabbar = UIManager.CreateTabBar(PoseMe.forceTargets.Select(x => x.name).ToArray(), SelectRegion, columns:8);
            if (PoseMe.currentPose == null)
            {
                PoseMe.singleton.CreateTextField(new JSONStorableString("bla", "Create a pose first and come back!"));
                return;
            }
            currentLimbIdle = PoseMe.currentPose.poseIdle.limbIdles[lastTabId];
            directionalForceUid = UIManager.CreateV3Slider(currentLimbIdle.directionalForce);
            
            
            forceOnesidedUid = UIManager.CreateToggleArray(currentLimbIdle.forceOnesided);
            forceOnesidedUid.RegisterBools(currentLimbIdle.forceOnesided);

            directionalTorqueUid = UIManager.CreateV3Slider(currentLimbIdle.directionalTorque);
            torqueOnesidedUid = UIManager.CreateToggleArray(currentLimbIdle.torqueOnesided);
            torqueOnesidedUid.RegisterBools(currentLimbIdle.torqueOnesided);
            
            // forceEnabledUid = PoseMe.singleton.CreateToggle(lastLimbIdle.forceEnabled);
            forceEnabledUid = UIManager.CreateReusableUIDynamicToggle();
            torqueEnabledUid = UIManager.CreateReusableUIDynamicToggle();
            forceEnabledUid.RegisterBool(currentLimbIdle.forceEnabled);
            torqueEnabledUid.RegisterBool(currentLimbIdle.torqueEnabled);
            
            currentLimbIdle.poseIdle.scale.CreateUI();
            idlesEnabledUid = UIManager.CreateReusableUIDynamicToggle();
            // torqueEnabledUid = PoseMe.singleton.CreateToggle(lastLimbIdle.torqueEnabled);
            PoseMe.singleton.CreateSpacer(true).ForceHeight(245);
            PoseMe.applyIdles.CreateUI(UIElements);
            
            PoseMe.singleton.CreateTextField(info, true);
            PoseMe.singleton.SetupButton("All Regions To Default", currentLimbIdle.poseIdle.ResetToDefault, PoseMe.warningColor, UIElements, true);
            PoseMe.singleton.SetupButton("Disable All Regions", currentLimbIdle.poseIdle.DisableAll, PoseMe.warningColor, UIElements, true);
            
            tabbar.SelectTab(lastTabId);
            uiOpen = true;
        }

        public static void SelectRegion(int id)
        {
            // PoseMe.currentPose.poseForce.limbForces[lastTabId].enabled.toggle
            
            // lastLimbIdle.forceEnabled.DeregisterToggle();
            // lastLimbIdle.torqueEnabled.DeregisterToggle();
            
            lastTabId = id;
            var lastPoseIdle = currentLimbIdle.poseIdle;
            currentLimbIdle = PoseMe.currentPose.poseIdle.limbIdles[id];

            forceEnabledUid.RegisterBool(currentLimbIdle.forceEnabled);
            torqueEnabledUid.RegisterBool(currentLimbIdle.torqueEnabled);
            idlesEnabledUid.RegisterBool(PoseMe.currentPose.poseIdle.applyIdles);
            directionalForceUid.RegisterVector(currentLimbIdle.directionalForce);
            directionalTorqueUid.RegisterVector(currentLimbIdle.directionalTorque);
            
            forceOnesidedUid.RegisterBools(currentLimbIdle.forceOnesided);
            torqueOnesidedUid.RegisterBools(currentLimbIdle.torqueOnesided);
            
            if(lastPoseIdle != currentLimbIdle.poseIdle)
            {
                PoseMe.singleton.RemoveSlider(lastPoseIdle.scale);
                PoseMe.currentPose.poseIdle.scale.CreateUI();
            }
            
            Debug.Clear();
            AxisSetActive(true);
            info.val = currentLimbIdle.GetInfo();
        }

        public static void AxisSetActive(bool val)
        {
            if(val) PoseMe.forceTargets[lastTabId].transform.Draw();
            else Debug.Clear();
            axisShown = val;
        }
    }
}