using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MacGruber;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using UnityEngine;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace CheesyFX
{
    public class Pose
    {
        public UIDynamicButton uiButton;
        private ClickListener clickListener;
        public JSONClass pose;
        public List<CamAngle> camAngles = new List<CamAngle>();
        private int _id;
        public static IEnumerator smoothCam;
        public EventTrigger onPoseEnter;
        public EventTrigger onPoseExit;
        private static Color selectedBGColor = new Color(0.48f, 0.48f, 0.48f);
        private static Color deselectedBGColor = new Color(0.76f, 0.76f, 0.76f, 0f);
        private static Color selectedColor = Color.white;
        public static Color deselectedColor = new Color(1f, 1f, 1f, PoseMe.buttonTransparency.val);
        private bool _selected;
        public CamAngle currentAngle;
        
        public Dictionary<Atom, JSONStorableBool> actorToggles = new Dictionary<Atom, JSONStorableBool>();

        public PoseIdle poseIdle;
        public string actors => "<b>Stored actors:</b> "+string.Join(", ", pose["pose"].AsObject.Keys.ToArray());
        public bool selected
        {
            get { return _selected;}
            set
            {
                _selected = value;
                clickListener.dragEnabled = value;
                if (value)
                {
                    uiButton.button.image.color = selectedColor;
                    backdropImage.color = selectedBGColor;
                    if(PoseMe.applyIdles.val && poseIdle.applyIdles.val) poseIdle.enabled = true;
                    poseIdle.ResetForces();
                }
                else
                {
                    uiButton.button.image.color = deselectedColor;
                    backdropImage.color = deselectedBGColor;
                    if(poseIdle) poseIdle.enabled = false;
                }
            }
        }

        private static Atom containingAtom => PoseMe.atom;
        public int id
        {
            get { return _id; }
            set
            {
                if (value == _id) return;
                _id = value;
                if(uiButton != null) uiButton.label = _id.ToString();
            }
        }

        public Pose(int id)
        {
            if(PoseMe.currentPose != null)
            {
                PoseMe.currentPose.selected = false;
            }
            _id = id;
            SetupActorToggles();
            onPoseEnter = new EventTrigger(PoseMe.singleton, "On Pose Enter");
            onPoseExit = new EventTrigger(PoseMe.singleton, "On Pose Exit");
            poseIdle = PoseMe.singleton.gameObject.AddComponent<PoseIdle>().Init(this);
            currentAngle = new CamAngle(this, (byte[])null);
            camAngles.Add(currentAngle);
            GetPose();
            PoseMe.currentPose = this;
            PoseMe.onPoseEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseEnterTriggerButton.button.onClick.AddListener(onPoseEnter.OpenPanel);
            PoseMe.onPoseExitTriggerButton.button.onClick.AddListener(onPoseExit.OpenPanel);
            AddButton();
            backdropImage.color = selectedBGColor;
            PoseMe.actorsJ.val = actors;
            selected = true;
            TakeScreenshot(UpdateCurrentAngle);
            clickListener.dragEnabled = true;
        }
        
        public Pose(int id, JSONClass pose)
        {
            this.pose = pose;
            _id = id;
            onPoseEnter = new EventTrigger(PoseMe.singleton, "On Pose Enter");
            onPoseExit = new EventTrigger(PoseMe.singleton, "On Pose Exit");
            poseIdle = PoseMe.singleton.gameObject.AddComponent<PoseIdle>().Init(this);
            // PoseMe.currentPose = this;
            PoseMe.onPoseEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseEnterTriggerButton.button.onClick.AddListener(onPoseEnter.OpenPanel);
            PoseMe.onPoseExitTriggerButton.button.onClick.AddListener(onPoseExit.OpenPanel);
            foreach (JSONClass jc in pose["pose"][containingAtom.name]["storables"].AsArray)
            {
                if (jc["id"].Value != "control") continue;
                storedWorldToLocalMatrix = Matrix4x4.TRS(jc["position"].AsObject.ToV3(), Quaternion.Euler(jc["rotation"].AsObject.ToV3()), Vector3.one).inverse;
                break;
            }
            onPoseEnter.RestoreFromJSON(pose, PoseMe.singleton.subScenePrefix, false, true);
            onPoseExit.RestoreFromJSON(pose, PoseMe.singleton.subScenePrefix, false, true);
            if (pose.HasKey("cams"))
            {
                foreach (var cam in pose["cams"].Childs)
                {
                    camAngles.Add(new CamAngle(this, cam.AsObject));
                }

                if(camAngles.Count > 0) currentAngle = camAngles.First();
            }
            else if (pose.HasKey("img"))
            {
                var camAngle = new CamAngle(this, Convert.FromBase64String(pose["img"]))
                {
                    hasCam = false
                };
                camAngles.Add(camAngle);
                currentAngle = camAngle;
            }
            else if (pose.HasKey("images"))
            {
                var camAngle = new CamAngle(this, Convert.FromBase64String(pose["images"].Childs.First()))
                {
                    hasCam = false
                };
                camAngles.Add(camAngle);
                currentAngle = camAngle;
            }
            if (pose.HasKey("idles"))
            {
                poseIdle.Load(pose["idles"].AsObject);
            }

            SetupActorToggles();
            AddButton();
            
            SetButtonImage();
        }

        public void Destroy()
        {
            Object.DestroyImmediate(poseIdle);
            RemoveButton();
            onPoseEnter.Remove();
            onPoseExit.Remove();
            foreach (var cam in camAngles)
            {
                cam.onCamEnter.Remove();
                cam.onCamExit.Remove();
            }
        }

        private void SetupActorToggles()
        {
            PoseMe.persons.ForEach(x => actorToggles[x.atom] = new JSONStorableBool(x.atom.uid, true));
            PoseMe.dildos.ForEach(x => actorToggles[x] = new JSONStorableBool(x.uid, true));
            PoseMe.toys.ForEach(x => actorToggles[x] = new JSONStorableBool(x.uid, true));
            PoseMe.singleton.RegisterActorToggles();
        }

        public void SyncActorToggleNames()
        {
            foreach (var item in actorToggles)
            {
                item.Value.name = item.Key.name;
            }
        }

        private void MapAtoms(JSONClass pose)
        {
            foreach (var person in FillMeUp.persons)
            {
                if(pose.HasKey(person.atom.uid)) continue;
                if (person.atom == containingAtom)
                {
                    var jc = pose.Childs.First(x => x.AsObject["containingAtom"].AsBool);
                }
            }
        }

        public void UpdatePose()
        {
            GetPose();
            PoseMe.actorsJ.val = actors;
        }
        
        public void AddCamAngle()
        {
            TakeScreenshot(AddCamAngle);
        }
        
        public void UpdateAngle()
        {
            TakeScreenshot(UpdateCurrentAngle);
        }

        public void DeleteAngle()
        {
            if (currentAngle == null) return;
            if (camAngles.Count == 1)
            {
                "A pose has to have at least one camera angle.".Print();
                return;
            }
            camAngles.Remove(currentAngle);
            PoseMe.SyncAngleChooser();
            // PoseMe.angleChooser.valNoCallback = "";
            // PoseMe.previewImage.sprite = null;
            currentAngle = camAngles[0];
            
            RenameAngles();
            currentAngle.Apply();
        }

        private void RenameAngles()
        {
            for (int i = 0; i < camAngles.Count; i++)
            {
                camAngles[i].id = i;
            }
        }
        
        public void UpdateJSON()
        {
            pose[onPoseEnter.Name] = onPoseEnter.GetJSON(PoseMe.singleton.subScenePrefix);
            pose[onPoseExit.Name] = onPoseExit.GetJSON(PoseMe.singleton.subScenePrefix);
            pose["idles"] = poseIdle.Store();
            JSONArray ja = new JSONArray();
            foreach (var camAngle in camAngles)
            {
                ja.Add(camAngle.Store());
            }
            pose["cams"] = ja;
        }
        
        private void GetPose()
        {
            storedWorldToLocalMatrix = containingAtom.mainController.transform.worldToLocalMatrix;
            refMatrix = containingAtom.mainController.transform.localToWorldMatrix;
            // PoseMe.dildos.Clear();
            var jc = new JSONClass();
            jc["SceneType"] = PoseMe.sceneType;
            jc["id"] = id.ToString();
            jc["idles"] = poseIdle.Store();
            var jcPose = jc["pose"] = new JSONClass();
            jcPose[containingAtom.uid] = GetPersonPose(containingAtom);
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                if (atom != containingAtom && atom.type == "Person") jcPose[atom.uid] = GetPersonPose(atom);
                else if(atom.IsToyOrDildo()) jcPose[atom.uid] = GetToyPose(atom);
            }
            if(PoseMe.manageForces.val)
            {
                var jcForces = new JSONClass();
                var hasForces = false;
                foreach (var fuckable in FillMeUp.fuckables)
                {
                    if (!fuckable.isPenetrated) continue;
                    var jcFuckable = jcForces[fuckable.name] = new JSONClass();
                    fuckable.StoreForces(jcFuckable.AsObject);
                    hasForces = true;
                }

                if (hasForces) jc["Forces"] = jcForces;
            }
            jc[onPoseEnter.Name] = onPoseEnter.GetJSON(PoseMe.singleton.subScenePrefix);
            jc[onPoseExit.Name] = onPoseExit.GetJSON(PoseMe.singleton.subScenePrefix);
            // foreach (var item in BodyRegionMapping.touchZones)
            // {
            //     if(item.Value.isTouched) $"{item.Value.name} {item.Value.isTouched}".Print();
            // }
            
            pose = jc;
            PoseMe.singleton.RegisterActorToggles();
        }

        private JSONClass GetPersonPose(Atom atom)
        {
            JSONClass poseJSON = null;
            var storables = atom.GetStorableIDs().Select(atom.GetStorableByID)
                .Where(t => !t.exclude && t.gameObject.activeInHierarchy)
                .Where(ShouldStorableBeIncluded);
            var storablesJSON = new JSONArray();
            
            foreach (var storable in storables)
            {
                var jc = storable.GetJSON(true, false, storable.name != "geometry");
                if (atom != containingAtom && jc["id"].Value == "control")
                {
                    jc["position"] = storedWorldToLocalMatrix.MultiplyPoint(atom.mainController.transform.position).ToJC();
                    jc["rotation"] = (storedWorldToLocalMatrix.rotation * atom.mainController.transform.rotation).eulerAngles.ToJC();
                }
                storablesJSON.Add(jc);
            }

            var genderListener = FillMeUp.persons.First(x => x.atom == atom).genderListener;
            poseJSON = new JSONClass
            {
                ["setUnlistedParamsToDefault"] = { AsBool = true },
                ["V2"] = { AsBool = true },
                ["type"] = atom.type,
                ["on"] = {AsBool = atom.GetBoolParamValue("on")},
                ["containingAtom"] = {AsBool = atom == PoseMe.atom},
                ["gender"] = genderListener.dcs.gender.ToString(),
                ["futa"] = {AsBool = genderListener.isFuta},
                ["handPose"] = GetHandPose(atom),
                ["storables"] = storablesJSON
            };
            return poseJSON;
        }
        
        private static bool ShouldStorableBeIncluded(JSONStorable t)
        {
            if (t is FreeControllerV3) return true;
            if (t is DAZBone) return true;
            if (t.storeId == "geometry" && t is DAZCharacterSelector)
            {
                // var morph = t as DAZMorph;
                return false;
            }
            return false;
        }

        private JSONArray GetHandPose(Atom atom)
        {
            var ja = new JSONArray();
            var dcs = atom.GetStorableByID("geometry") as DAZCharacterSelector;
            List<JSONClass> morphs = new List<JSONClass>();
            foreach (var morph in dcs.morphsControlUI.GetMorphs())
            {
                if (!morph.isPoseControl || !morph.hasFormulas || !HasFingerFormula(morph)) continue;
                var jc = new JSONClass();
                if (morph.StoreJSON(jc, true))
                {
                    if (IsControlMorph(morph) && morph.morphValue == morph.jsonFloat.defaultVal)
                    {
                        morphs.Insert(0, jc);
                    }
                    else morphs.Add(jc);
                }
            }
            for (int i = 0; i < morphs.Count; i++)
            {
                ja.Add(morphs[i]);
            }
            return ja;
        }
        
        private bool HasFingerFormula(DAZMorph morph)
        {
            var validFormulas = morph.formulas.Any(x =>
                    (x.target.Contains("Thumb") ||
                    x.target.Contains("Index") ||
                    x.target.Contains("Mid") ||
                    x.target.Contains("Ring") ||
                    x.target.Contains("Pinky") ||
                    x.target.Contains("Finger")) &&
                    Mathf.Abs(x.multiplier) > .1f);
            return validFormulas;
        }

        private bool IsControlMorph(DAZMorph morph)
        {
            return morph.formulas.Any(x => x.target.StartsWith("CTRL"));
        }
        
        private JSONClass GetToyPose(Atom atom)
        {
            JSONClass jc = new JSONClass();
            jc["type"] = atom.type;
            jc["on"].AsBool = atom.GetBoolParamValue("on");
            jc["ctrl"] = atom.mainController.GetJSON();
            jc["ctrl"]["position"] = storedWorldToLocalMatrix.MultiplyPoint(atom.mainController.transform.position).ToJC();
            jc["ctrl"]["rotation"] = (storedWorldToLocalMatrix.rotation * atom.mainController.transform.rotation).eulerAngles.ToJC();
            foreach (var rb in atom.GetComponentsInChildren<Rigidbody>(true))
            {
                jc[rb.name] = new JSONClass
                {
                    ["pos"] = storedWorldToLocalMatrix.MultiplyPoint(rb.position).ToJA(),
                    ["rot"] = (storedWorldToLocalMatrix.rotation * rb.rotation).eulerAngles.ToJA()
                };
            }
            return jc;
        }

        public Matrix4x4 storedWorldToLocalMatrix;
        private Matrix4x4 currentLocalToWorldMatrix;
        public Matrix4x4 refMatrix;

        private int applyClickCount;

        public void Apply(bool applyCam = true)
        {
            PoseMe.SyncAngleChooser();
            PoseMe.actorsJ.val = actors;
            if (PoseMe.currentPose != null && PoseMe.currentPose != this)
            {
                PoseMe.currentPose.onPoseExit.Trigger();
                PoseMe.currentPose.selected = false;
            }
            onPoseEnter.Trigger();
            PoseMe.currentPose = this;
            PoseMe.onPoseEnterTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseExitTriggerButton.button.onClick.RemoveAllListeners();
            PoseMe.onPoseEnterTriggerButton.button.onClick.AddListener(onPoseEnter.OpenPanel);
            PoseMe.onPoseExitTriggerButton.button.onClick.AddListener(onPoseExit.OpenPanel);
            PoseMe.poseChooser.valNoCallback = PoseMe.poses.IndexOf(this).ToString();
            PoseMe.applyPosePost.Stop();
            PoseMe.softPhysicsStates.Clear();
            for (int i = 0; i < FillMeUp.persons.Count; i++)
            {
                PoseMe.softPhysicsStates.Add(PoseMe.persons[i].GetSoftPhysicsState());
            }
            currentLocalToWorldMatrix = containingAtom.mainController.transform.localToWorldMatrix;
            refMatrix = !PoseMe.restoreRoot.val ? currentLocalToWorldMatrix : storedWorldToLocalMatrix.inverse;
            ApplyPersonPose(containingAtom);
            foreach (string atomUid in pose["pose"].AsObject.Keys)
            {
                if(atomUid == PoseMe.atom.uid) continue;
                var atom = SuperController.singleton.GetAtomByUid(atomUid);
                if (atom == null)
                {
                    $"Atom {atomUid} not found".Print();
                    continue;
                }
                if(atom.type == "Person")
                {
                    ApplyPersonPose(atom);
                }
            }
            
            PoseMe.applyPosePost = ApplyPosePost(pose.AsObject, applyCam).Start();
            selected = true;
            if(IdleUIProvider.uiOpen) IdleUIProvider.SelectRegion(IdleUIProvider.lastTabId);
            // if(randomizeAngles != null) PoseMe.poseRandomizeCamUid.RegisterBool(randomizeAngles);
            clickListener.dragEnabled = true;
            PoseMe.SyncAngleChooser();
            PoseMe.singleton.RegisterActorToggles();
        }

        private void ApplyPersonPose(Atom atom)
        {
            try
            {
                if(!actorToggles[atom].val) return;
                Quaternion headCtrlRot = Quaternion.identity;
                Quaternion headRot = Quaternion.identity;
                Quaternion neckRot = Quaternion.identity;
                Quaternion chestRot = Quaternion.identity;
                FreeControllerV3 headCtrl = null;
                DAZBone head = null;
                DAZBone neck = null;
                DAZBone chest = null;
                
                var personJc = pose["pose"][atom.uid].AsObject;
                if(personJc.HasKey("on")) atom.SetBoolParamValue("on", personJc["on"].AsBool);
                // if(!atom.GetBoolParamValue("on")) return;
            
                if (!PoseMe.restoreHeadRotation.val)
                {
                    headCtrl = atom.freeControllers.First(x => x.name == "headControl");
                    var bones = atom.GetComponentsInChildren<DAZBone>();
                    for (int i = 0; i < bones.Length; i++)
                    {
                        var bone = bones[i];
                        if (bone.name == "head") head = bone;
                        else if (bone.name == "neck") neck = bone;
                        else if (bone.name == "chest") chest = bone;
                    }
                    headCtrlRot = headCtrl.transform.localRotation;
                    headRot = head.transform.localRotation;
                    neckRot = neck.transform.localRotation;
                    chestRot = chest.transform.localRotation;
                }
            
                PresetManagerControl pmc = atom.presetManagerControls.First(x => x.name == "PosePresets");
                var pm = pmc.GetComponent<PresetManager>();
                
                if(atom == containingAtom)
                {
                    
                    if (!PoseMe.restoreRoot.val)
                    {
                        
                        var controlJc = personJc["storables"].Childs.First(x => x["id"].Value == "control");
                        controlJc["id"] = "_control";
                        pm.LoadPresetFromJSON(personJc);
                        controlJc["id"] = "control";
                    }
                    else pm.LoadPresetFromJSON(personJc);
                }
                else
                {
                    var controlJc = personJc["storables"].Childs.First(x => x["id"].Value == "control");
                    var storedPos = controlJc["position"].AsObject.ToV3();
                    var storedRot = controlJc["rotation"].AsObject.ToV3();
                    
                    controlJc["position"] = refMatrix.MultiplyPoint(storedPos).ToJC();
                    controlJc["rotation"] = (refMatrix.rotation * Quaternion.Euler(storedRot)).eulerAngles.ToJC();
                    pm.LoadPresetFromJSON(personJc);
                    controlJc["position"] = storedPos.ToJC();
                    controlJc["rotation"] = storedRot.ToJC();
                }

                if (!PoseMe.restoreHeadRotation.val)
                {
                    headCtrl.transform.localRotation = headCtrlRot;
                    head.transform.localRotation = headRot;
                    neck.transform.localRotation = neckRot;
                    chest.transform.localRotation = chestRot;
                }

                if (personJc.HasKey("handPose"))
                {
                    var dcs = atom.GetStorableByID("geometry") as DAZCharacterSelector;
                    foreach (var morphJc in personJc["handPose"].Childs)
                    {
                        var morph = dcs.morphsControlUI.GetMorphByUid(morphJc["uid"]);
                        morph.morphValue = morphJc["value"].AsFloat;
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
        }
        
        private IEnumerator ApplyDildoPose(Atom atom, JSONClass pose)
        {
            if(pose.HasKey("on")) atom.SetBoolParamValue("on", pose["on"].AsBool);
            yield return new WaitForFixedUpdate();
            // if(!atom.GetBoolParamValue("on")) return;
            var controlJc = pose["control"];
            var pos = controlJc["pos"].AsArray.ToV3();
            var rot = controlJc["rot"].AsArray.ToV3();
            
            atom.mainController.transform.position = refMatrix.MultiplyPoint(pos);
            atom.mainController.transform.rotation = refMatrix.rotation * Quaternion.Euler(rot);
            
            foreach (var rb in atom.GetComponentsInChildren<Rigidbody>())
            {
                rb.position = refMatrix.MultiplyPoint(pose[rb.name]["pos"].AsArray.ToV3());
                rb.rotation = refMatrix.rotation * Quaternion.Euler(pose[rb.name]["rot"].AsArray.ToV3());
            }
        }

        private IEnumerator ApplyToyPose(Atom atom, JSONClass pose)
        {
            if(pose.HasKey("on")) atom.SetBoolParamValue("on", pose["on"].AsBool);
            yield return new WaitForFixedUpdate();
            if(!pose.HasKey("ctrl")) yield break;
            var controlJc = pose["ctrl"].AsObject;
            atom.mainController.RestoreFromJSON(controlJc);
            var pos = controlJc["position"].AsObject.ToV3();
            var rot = controlJc["rotation"].AsObject.ToV3();
            
            atom.mainController.transform.position = refMatrix.MultiplyPoint(pos);
            atom.mainController.transform.rotation = refMatrix.rotation * Quaternion.Euler(rot);
            foreach (var rb in atom.GetComponentsInChildren<Rigidbody>())
            {
                rb.position = refMatrix.MultiplyPoint(pose[rb.name]["pos"].AsArray.ToV3());
                rb.rotation = refMatrix.rotation * Quaternion.Euler(pose[rb.name]["rot"].AsArray.ToV3());
            }
        }

        private static List<Atom> toysToRestore = new List<Atom>();
        private static WaitForFixedUpdate wait = new WaitForFixedUpdate();

        private bool ShouldAtomBeRestored(Atom atom, JSONClass poseJc)
        {
            bool typeIncluded;
            if (atom.type == "Person") typeIncluded = atom == containingAtom || PoseMe.restoreOtherPersons.val;
            else typeIncluded = atom.type == "Dildo" ? PoseMe.restoreDildos.val : PoseMe.restoreToys.val;
            if (!typeIncluded) return false;
            JSONStorableBool bJson;
            return actorToggles.TryGetValue(atom, out bJson) && bJson.val && poseJc.HasKey(atom.uid);
        }
        
        private bool ShouldAtomBeRestored(Atom atom)
        {
            bool typeIncluded;
            if (atom.type == "Person") typeIncluded = atom == containingAtom || PoseMe.restoreOtherPersons.val;
            else typeIncluded = atom.type == "Dildo" ? PoseMe.restoreDildos.val : PoseMe.restoreToys.val;
            if (!typeIncluded) return false;
            JSONStorableBool bJson;
            return actorToggles.TryGetValue(atom, out bJson) && bJson.val;
        }
        
        private IEnumerator ApplyPosePost(JSONClass pose, bool applyCam)
        {
            var poseJc = pose["pose"].AsObject;
            toysToRestore.Clear();
            for (int i = 0; i < PoseMe.dildos.Count; i++)
            {
                if(ShouldAtomBeRestored(PoseMe.dildos[i], poseJc)) toysToRestore.Add(PoseMe.dildos[i]);
            }
            for (int i = 0; i < PoseMe.toys.Count; i++)
            {
                if(ShouldAtomBeRestored(PoseMe.toys[i], poseJc)) toysToRestore.Add(PoseMe.toys[i]);
            }

            for (int i = 0; i < toysToRestore.Count; i++)
            {
                // toysToRestore[i].collisionEnabled = false;
                toysToRestore[i].SetBoolParamValue("on", false);
            }

            
            if (applyCam && camAngles.Count > 0)
            {
                yield return wait;
                yield return wait;
                yield return wait;
                smoothCam.Stop();
                currentAngle.Apply();
                // ApplyNextCamAngle();
            
            }
            yield return new WaitUntil(() =>
            {
                for (int i = 0; i < PoseMe.persons.Count; i++)
                {
                    if (PoseMe.persons[i].GetSoftPhysicsState() != PoseMe.softPhysicsStates[i]) return false;
                }
                return true;
            });
            
            foreach (string atomUid in poseJc.Keys)
            {
                var atom = SuperController.singleton.GetAtomByUid(atomUid);
                if (atom == null) continue;
                var atomPose = poseJc[atomUid].AsObject;
                if (atom.type == "Person")
                {
                    // JSONArray storables = atomPose["storables"].AsArray;
                    // foreach (var bone in atom.GetComponentsInChildren<DAZBone>())
                    // {
                    //     
                    //     for (int i = 0; i < storables.Count; i++)
                    //     {
                    //         JSONClass jc = storables[i].AsObject;
                    //         if(jc["id"] != bone.name) continue;
                    //         bone.Print();
                    //         bone.RestoreFromJSON(jc);
                    //     }
                    // }
                }
                else if (ShouldAtomBeRestored(atom)) ApplyToyPose(atom, atomPose).Start();
            }

            if (PoseMe.manageForces.val && pose.HasKey("Forces"))
            {
                var jcForces = pose["Forces"].AsObject;
                foreach (var fuckable in FillMeUp.fuckables)
                {
                    if(!jcForces.HasKey(fuckable.name)) continue;
                    fuckable.LoadForces(jcForces[fuckable.name].AsObject);
                }
            }
        }

        public void ApplyNextCamAngle()
        {
            if (camAngles.Count <= 1)
            {
                Apply();
                return;
            }
            camAngles[(currentAngle.id+1) % camAngles.Count].Apply();
        }

        public void ApplyPreviousCamAngle()
        {
            if (camAngles.Count <= 1)
            {
                Apply();
                return;
            }
            int id = currentAngle.id - 1;
            if (id == -1) id = camAngles.Count - 1;
            camAngles[id].Apply();
        }

        public void ApplyRandomCamAngle()
        {
            if (camAngles.Count <= 1)
            {
                Apply();
                return;
            }
            camAngles.TakeRandom(currentAngle.id).Apply();
        }

        public RectTransform buttonRT;
        private Image backdropImage;
        
        private void AddButton()
        {
            try
            {
                buttonRT = Object.Instantiate(PoseMe.singleton.manager.configurableButtonPrefab) as RectTransform;
                Object.Destroy(buttonRT.GetComponentInChildren<UIDynamicButton>());
                Object.Destroy(buttonRT.GetComponentInChildren<Button>());
                Object.Destroy(buttonRT.GetComponentInChildren<Text>());
                backdropImage = buttonRT.GetComponent<Image>();
                backdropImage.color = deselectedBGColor;
                buttonRT.SetParent(PoseMe.canvas.transform, false);
            
                RectTransform button = Object.Instantiate(PoseMe.singleton.manager.configurableButtonPrefab) as RectTransform;
                // button.SetParent(PoseMe.canvas.transform, false);
                button.SetParent(buttonRT, false);
                button.anchorMax = new Vector2(.98f, .98f);
                button.anchorMin = new Vector2(.02f, .02f);
                button.offsetMax = new Vector2();
                button.offsetMin = new Vector2();

                uiButton = button.GetComponent<UIDynamicButton>();
                uiButton.label = _id.ToString();
                uiButton.buttonText.fontSize = 25;
                uiButton.buttonText.color = Color.white;
                uiButton.button.image.color = deselectedColor;
                // uiButtonTextAlt.alignment = TextAnchor.UpperRight;
                // uiButton.button.onClick.AddListener(ApplyWaitForDoubleClick);
                var cb = uiButton.button.colors;
                cb.pressedColor = cb.highlightedColor = cb.normalColor;
                uiButton.button.colors = cb;
                clickListener = button.gameObject.AddComponent<ClickListener>();
                clickListener.onLeftClick.AddListener(() =>
                {
                    if (PoseMe.currentPose == this) PoseMe.onLeftClick.Invoke();
                    else
                    {
                        Apply();
                        if(PoseMe.leftClickChooser.val.StartsWith("R")) ApplyRandomCamAngle();
                    }
                });
                clickListener.onRightClick.AddListener(() =>
                {
                    if (PoseMe.currentPose == this) PoseMe.onRightClick.Invoke();
                    else
                    {
                        Apply();
                        if(PoseMe.rightClickChooser.val.StartsWith("R")) ApplyRandomCamAngle();
                    }
                });
                clickListener.onMiddleClick.AddListener(() => Apply(false));
                clickListener.onPointerEnter.AddListener(() => {
                    GlobalSceneOptions.singleton.disableNavigation = true;
                    SuperController.singleton.disableNavigation = true;
                    PoseMe.buttonHovered = true;
                    PoseMe.PrintInputConfig();
                });
                clickListener.onPointerExit.AddListener(() => {
                    GlobalSceneOptions.singleton.disableNavigation = false;
                    SuperController.singleton.disableNavigation = false;
                    PoseMe.buttonHovered = false;
                });
                clickListener.onDragUp.AddListener(PoseMe.onDragUp);
                clickListener.onDragDown.AddListener(PoseMe.onDragDown);
                clickListener.onDragLeft.AddListener(PoseMe.onDragLeft);
                clickListener.onDragRight.AddListener(PoseMe.onDragRight);
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
        }

        public void RemoveButton()
        {
            Object.Destroy(uiButton.transform.parent.gameObject);
        }

        public void StoreImage(string jsonPath)
        {
            if(camAngles.Count > 0) FileManagerSecure.WriteAllBytes(jsonPath.Replace(".json", ".jpg"), camAngles[0].img);
        }

        // public void DeleteFromDisk()
        // {
        //     if (!FileManagerSecure.FileExists(path.val + ".json")) return;
        //     FileManagerSecure.DeleteFile(path.val+".json");
        //     FileManagerSecure.DeleteFile(path.val+".jpg");
        // }

        public void StoreBackup()
        {
            var backupPath = $"{PoseMe.backUpStorePath}{PoseMe.sceneType}_{DateTime.Now:yyyy-MM-dd-HHMMss}";
            if(camAngles.Count > 0) FileManagerSecure.WriteAllBytes(backupPath+".jpg", camAngles[0].img);
            SuperController.singleton.SaveJSON(pose, backupPath+".json");
            // FileManagerSecure.MoveFile(path.val+".json", backupName + ".json");
            // FileManagerSecure.MoveFile(path.val + ".jpg", backupName + ".jpg");
            $"PoseMe: Backup stored at {backupPath}.".Print();
        }

        public void SyncAtomNames(string oldUid, string newUid)
        {
            onPoseExit.SyncAtomNames();
            onPoseExit.SyncAtomNames();
            var atomPoses = pose["pose"];
            atomPoses[newUid] = atomPoses[oldUid];
            atomPoses.Remove(oldUid);
        }

        private void TakeScreenshot(Action<byte[]> callback)
        {
            SuperController.singleton.DoSaveScreenshot(PoseMe.tmpPath+"1.jpg", val =>
            {
                callback(GetImageBytes(val));
            });
        }

        private byte[] GetImageBytes(string imagePath)
        {
            // if(!FileManagerSecure.FileExists(imagePath)) return null;
            var image = FileManagerSecure.ReadAllBytes(imagePath);
            FileManagerSecure.DeleteFile(imagePath);
            return image;
        }

        private void UpdateCurrentAngle(byte[] image)
        {
            currentAngle.Update(image);
            SetButtonImage();
            SetUIImage(image);
        }

        private void AddCamAngle(byte[] image)
        {
            currentAngle = new CamAngle(this, image);
            camAngles.Add(currentAngle);
            SetButtonImage();
            SetUIImage(image);
            var choices = PoseMe.camChooser.choices;
            PoseMe.camChooser.choices = null;
            choices.Add(currentAngle.id.ToString());
            PoseMe.camChooser.choices = choices;
            PoseMe.camChooser.val = currentAngle.id.ToString();
        }

        public void MoveAngle(bool up)
        {
            if(currentAngle == null) return;
            int i = currentAngle.id;
            if (up)
            {
                if (i == 0) return;
                var upperPose = camAngles.LastOrDefault(x => x.id < i);
                if (upperPose != null)
                {
                    camAngles[i-1] = currentAngle;
                    camAngles[i] = upperPose;
                    upperPose.id++;
                    currentAngle.id--;
                }

                PoseMe.camChooser.valNoCallback = (i - 1).ToString();
            }
            else
            {
                if(i == camAngles.Count - 1) return;
                var lowerPose = camAngles.FirstOrDefault(x => x.id > i);
                if (lowerPose != null)
                {
                    camAngles[i+1] = currentAngle;
                    camAngles[i] = lowerPose;
                    lowerPose.id--;
                    currentAngle.id++;
                }
            }
            SetButtonImage();
            PoseMe.camChooser.valNoCallback = currentAngle.id.ToString();
        }
        
        // public void FetchThumbnail(string path)
        // {
        //     img = FileManagerSecure.ReadAllBytes(path);
        //     FileManagerSecure.WriteAllBytes(this.path.val+".jpg", img);
        //     SetButtonImage(PoseMe.showThumbnails.val);
        // }
        
        public void SetButtonImage()
        {
            if(PoseMe.showThumbnails.val && camAngles.Count > 0)
            {
                var texture = LoadTexture(camAngles[0].img);
                uiButton.button.image.sprite = Sprite.Create(texture,
                    new Rect(0.0f, 0.0f, texture.width, texture.height), new Vector2(0.5f, 0.5f));
            }
            else
            {
                uiButton.button.image.sprite = null;
            }
            SetButtonText();
        }

        public void SetButtonText()
        {
            if (PoseMe.showThumbnails.val)
            {
                if(camAngles.Count > 1)
                {
                    uiButton.buttonText.alignment = TextAnchor.LowerRight;
                    uiButton.buttonText.color = Color.white;
                    uiButton.buttonText.fontSize = 25;
                    // uiButton.buttonText.text = currentAngle.id.ToString();
                    uiButton.buttonText.text = $"{currentAngle.id} / {camAngles.Count-1}";
                }
                else uiButton.buttonText.color = Color.clear;
            }
            else
            {
                uiButton.buttonText.alignment = TextAnchor.MiddleCenter;
                uiButton.buttonText.color = Color.black;
                if(PoseMe.buttonSizeJ.val >= PoseMe.buttonSizeJ.defaultVal) uiButton.buttonText.fontSize = 30;
                else uiButton.buttonText.fontSize = (int)(30 * PoseMe.buttonSizeJ.val / PoseMe.buttonSizeJ.defaultVal);
                uiButton.buttonText.text = camAngles.Count > 1 ? $"{id} / {currentAngle.id}" : id.ToString();
            }
        }

        public void SetUIImage(byte[] image = null)
        {
            if(camAngles.Count == 0) return;
            if (image == null) image = camAngles[0].img;
            var texture = LoadTexture(image);
            if(PoseMe.tabbar.id == 0 && !IdleUIProvider.uiOpen)
            {
                PoseMe.UIImage.sprite = Sprite.Create(texture, new Rect(0.0f, 0.0f, texture.width, texture.height), new Vector2(0.5f, 0.5f));
                PoseMe.UIImageButton.buttonText.text = $"{currentAngle.id} / {camAngles.Count-1}";
            }
        }

        public static Texture2D LoadTexture(string path)
        {
            var tex = new Texture2D(2, 2);
            var data = FileManagerSecure.ReadAllBytes(path);
            tex.LoadImage(data);
            
            // var pixels = tex.GetPixels();
            // for (int i = 0; i < pixels.Length; i++)
            // {
            //     var pixel = pixels[i];
            //     
            //     if (pixel.ToHSV().V < .1f)
            //     {
            //         // pixel.ToHSV().V.Print();
            //         // pixel.a = 0f;
            //         // pixels[i] = pixel;
            //         var col = Color.white;
            //         col.a = 0f;
            //         pixels[i] = col;
            //     }
            //     // pixels[i] *= Color.Lerp(color, pixels[i], maskPixels[i].a);
            // }
            // tex.SetPixels(pixels);
            tex.Apply();
            return tex;
        }

        private static Texture2D LoadTexture(byte[] image)
        {
            var tex = new Texture2D(2, 2);
            tex.LoadImage(image);
            tex.Apply();
            return tex;
        }
    }
}