using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class PoseIdle : MonoBehaviour
    {
        public List<LimbIdle> limbIdles = new List<LimbIdle>();
        public JSONStorableBool applyIdles = new JSONStorableBool("Apply Idles (This Pose)", false);
        public JSONStorableFloat scale = new JSONStorableFloat("Scale (All Regions)", 1f, 0f, 5f, false);
        private Pose pose;
        
        public PoseIdle Init(Pose pose)
        {
            this.pose = pose;
            PoseMe.forceTargets.ForEach(x => limbIdles.Add(new LimbIdle(this, x)));
            applyIdles.setCallbackFunction += val =>
            {
                if (PoseMe.currentPose != pose) return;
                enabled = val && PoseMe.applyIdles.val;
            };
            // $"{pose.id} init".Print();
            enabled = false;
            
            return this;
        }

        private void OnEnable()
        {
            if(limbIdles.Count == 0) return;
            ResetForces();
        }
        //
        // private void OnDisable()
        // {
        //     $"{pose.id} disable".Print();
        // }

        // private void OnDestroy()
        // {
        //     $"{pose.id} destroy".Print();
        // }

        private void FixedUpdate()
        {
            if (SuperController.singleton.freezeAnimation) return;
            for (int i = 0; i < limbIdles.Count; i++)
            {
                limbIdles[i].Update();
            }
        }

        public void ResetForces()
        {
            for (int i = 0; i < limbIdles.Count; i++)
            {
                limbIdles[i].Reset();
            }
        }

        public void ResetToDefault()
        {
            limbIdles.ForEach(x => x.ResetToDefault());
        }
        
        public void DisableAll()
        {
            limbIdles.ForEach(x => x.forceEnabled.val = x.torqueEnabled.val = false);
        }

        public JSONClass Store()
        {
            var jc = new JSONClass();
            scale.Store(jc);
            applyIdles.Store(jc);
            foreach (var limbForce in limbIdles)
            {
                jc[limbForce.target.name] = limbForce.Store();
            }

            return jc;
        }

        public void Load(JSONClass jc)
        {
            scale.Load(jc);
            applyIdles.Load(jc);
            foreach (var limbForce in limbIdles)
            {
                limbForce.Load(jc);
            }             
        }
    }
}