using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MacGruber;
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;

namespace CheesyFX
{
    public class PoseMe : MVRScript
    {
        public static PoseMe singleton;
        private bool initialized;
        public static string libraryPath = $"Saves/PluginData/CheesyFX/BodyLanguage/PoseMe/Library/";
        public static string backUpStorePath = $"Saves/PluginData/CheesyFX/BodyLanguage/PoseMe/Backups/";
        public static string tmpPath = "Saves/PluginData/CheesyFX/BodyLanguage/PoseMe/tmp/";
        public static Pose currentPose;
        // private static string scenename;

        public static JSONStorableBool manageForces = new JSONStorableBool("Store & Restore Forces", true);
        public static JSONStorableBool restoreRoot = new JSONStorableBool("Restore Root", true);
        public static JSONStorableBool restoreHeadRotation = new JSONStorableBool("Restore Head Rotation", true);
        public static JSONStorableBool applyIdles = new JSONStorableBool("Apply Idles (Global)", false, val =>
        {
            if(currentPose == null) return;
            currentPose.poseIdle.enabled = val && currentPose.poseIdle.applyIdles.val;
        });
        
        private static JSONStorableBool printInputConfigOnFirstHover =
            new JSONStorableBool("Print Input Config On First Hover", true);
        
        public static JSONStorableStringChooser leftClickChooser =
            new JSONStorableStringChooser("leftClickChooser", new List<string>{"Apply Pose", "Next Cam", "Previous Cam", "Random Cam", "None"}, "Next Cam", "Left Click", val => SyncInput(val, ref onLeftClick));
        public static JSONStorableStringChooser rightClickChooser =
            new JSONStorableStringChooser("rightClickChooser", new List<string>{"Apply Pose", "Next Cam", "Previous Cam", "Random Cam", "None"}, "Previous Cam", "Right Click", val => SyncInput(val, ref onRightClick));
        public static JSONStorableStringChooser middleClickChooser =
            new JSONStorableStringChooser("middleClickChooser", new List<string>{"Apply Pose", "Next Cam", "Previous Cam", "Random Cam", "None"}, "Apply Pose", "Middle Click", val => SyncInput(val, ref onMiddleClick));
        public static JSONStorableStringChooser dragUpChooser =
            new JSONStorableStringChooser("dragUpChooser", new List<string>{"Apply Pose", "Next Cam", "Previous Cam", "Random Cam", "None"}, "Previous Cam", "Drag Up", val => SyncInput(val, ref onDragUp));
        public static JSONStorableStringChooser dragDownChooser =
            new JSONStorableStringChooser("dragDownChooser", new List<string>{"Apply Pose", "Next Cam", "Previous Cam", "Random Cam", "None"}, "Next Cam", "Drag Down", val => SyncInput(val, ref onDragDown));
        public static JSONStorableStringChooser dragLeftChooser =
            new JSONStorableStringChooser("dragLeftChooser", new List<string>{"Apply Pose", "Next Cam", "Previous Cam", "Random Cam", "None"}, "Random Cam", "Drag Left", val => SyncInput(val, ref onDragLeft));
        public static JSONStorableStringChooser dragRightChooser =
            new JSONStorableStringChooser("drarRightChooser", new List<string>{"Apply Pose", "Next Cam", "Previous Cam", "Random Cam", "None"}, "Random Cam", "Drag Right", val => SyncInput(val, ref onDragRight));
        
        public static JSONStorableStringChooser camMode = new JSONStorableStringChooser("camMode", new List<string>{"None", "Bezier", "Exponential", "Linear", "Snap"}, "Bezier", "Cam Mode");
        public static JSONStorableFloat smoothCamSpeedJ = new JSONStorableFloat("Smooth Cam Speed", 1f, .2f, 4f, true);
        public static JSONStorableFloat smoothCamBezierStrength = new JSONStorableFloat("Smooth Cam Bezier Strength", 1f, 0f, 5f, false);
        // public static JSONStorableFloat doubleClickTimeout = new JSONStorableFloat("Double Click Timeout", .3f, 0f, 1f, true);
        
        private static JSONStorableBool autoBackUpDeleted = new JSONStorableBool("Auto BackUp Deleted", false);
        public static JSONStorableBool showButtons = new JSONStorableBool("Show Buttons", true, val => canvas.gameObject.SetActive(val));
        public static JSONStorableBool showThumbnails = new JSONStorableBool("Show Thumbnails", true, OnShowThumbnailChanged);
        public static JSONStorableBool onlyPreviousAndNext = new JSONStorableBool("Only Show Previous & Next", false, (bool val) => LayoutButtons());
        public static MyUIDynamicToggle poseApplyIdlesUid;

        public static JSONStorableBool cinematicEnabled = new JSONStorableBool("Cinematic Enabled", false);
        public static int cinematicCamMode = 2;
        private static int cinematicPoseMode = 0;
        public static JSONStorableStringChooser cinematicCamModeChooser =
            new JSONStorableStringChooser("CinematicCamMode", new List<string>{"None", "Cycle Cams", "Shuffle Cams"}, "Shuffle Cams", "Cinematic Cam", val => cinematicCamMode = cinematicCamModeChooser.choices.IndexOf(val));
        public static JSONStorableStringChooser cinematicPoseModeChooser =
            new JSONStorableStringChooser("CinematicPoseMode", new List<string>{"None", "Cycle Poses", "Shuffle Poses"}, "None", "Cinematic Pose", val => cinematicPoseMode = cinematicPoseModeChooser.choices.IndexOf(val));
        private static JSONStorableFloat poseTimeMean = new JSONStorableFloat("Pose Time Mean", 30f, 1f, 30f);
        private static JSONStorableFloat poseTimeDelta = new JSONStorableFloat("Pose Time Delta", 10f, 0f, 30f);
        private static JSONStorableFloat camTimeMean = new JSONStorableFloat("Cam Time Mean", 10f, 2f, 30f);
        private static JSONStorableFloat camTimeDelta = new JSONStorableFloat("Cam Time Delta", 5f, 0f, 30f);
        private static JSONStorableFloat camSpeedMean = new JSONStorableFloat("Cam Speed Mean", .5f, .3f, 2f);
        private static JSONStorableFloat camSpeedDelta = new JSONStorableFloat("Cam Speed Delta", .3f, 0f, 1f);
        public static JSONStorableBool randomizeCamSpeed = new JSONStorableBool("Random Cam Speed", true);

        public static JSONStorableBool restoreOtherPersons = new JSONStorableBool("Restore Other Persons", true);
        public static JSONStorableBool restoreDildos = new JSONStorableBool("Restore Dildos", true);
        public static JSONStorableBool restoreToys = new JSONStorableBool("Restore BP & AH", true);
        
        public static List<Pose> poses = new List<Pose>();

        public static JSONStorableStringChooser poseChooser =
            new JSONStorableStringChooser("Apply Pose", new List<string>(), "", "Pose", ApplyPose);
        public static JSONStorableStringChooser camChooser =
            new JSONStorableStringChooser("Apply Angle", new List<string>{"0"}, "0", "Cam", ApplyCam);
        private static JSONStorableAction reapplyCurrent;
        private static JSONStorableAction nextPose;
        private static JSONStorableAction previousPose;
        private static JSONStorableAction randomPose;
        private static JSONStorableAction previousCam;
        private static JSONStorableAction nextCam;
        private static JSONStorableAction randomCam;
        private static JSONStorableAction printInputConfig;

        private Transform confirmCanvas;
        public static List<Atom> dildos = new List<Atom>();

        public static string sceneType;

        public static IEnumerator applyPosePost;
        public static List<int> softPhysicsStates = new List<int>();
        public static List<Person> persons => FillMeUp.persons;
        private UnityEventsListener uiListener;

        private float screenWidth;
        private float screenHeight;

        private static UIDynamicLabelInput sceneNameInput;

        private IOSystem ioSystem = new IOSystem();
        public static Atom atom;

        public static UIDynamicButton onPoseEnterTriggerButton;
        public static UIDynamicButton onPoseExitTriggerButton;
        public static UIDynamicButton onAngleEnterTriggerButton;
        public static UIDynamicButton onAngleExitTriggerButton;
        
        
        public static GameObject canvas;
        private JSONStorableFloat posX = new JSONStorableFloat("Pos X", 1f, 0f, 1f);
        private JSONStorableFloat posY = new JSONStorableFloat("Pos Y", 1f, 0f, 1f);
        public static JSONStorableFloat buttonSizeJ = new JSONStorableFloat("Button Size", 200f, SyncButtonSize,0f, 1000f);
        public static JSONStorableFloat buttonTransparency = new JSONStorableFloat("Button Transparency", .5f, val =>
        {
            Pose.deselectedColor.a = val;
            poses.ForEach(x => x.selected = x.selected);
        }, 0f, 1f);
        private static float buttonSize;
        public static JSONStorableFloat maxRows = new JSONStorableFloat("MaxRows", 10f, (float val) => LayoutButtons(), 1f, 50f, false);
        private static float buttonWidth = buttonSizeJ.defaultVal;
        private static float buttonHeight = buttonSizeJ.defaultVal;
        private static float buttonSpacing = 20f;

        public static Image UIImage;
        public static UIDynamicButton UIImageButton;
        public static JSONStorableString actorsJ = new JSONStorableString("actors", "");

        public static List<Rigidbody> forceTargets = new List<Rigidbody>();

        public static Transform cameraRig;
        public static Transform camera;
        public static bool isVR;

        public static List<object> UIElements = new List<object>();
        public static UIDynamicTabBar tabbar;
        public int lastTabId;
        
        public Dictionary<Atom, MyUIDynamicToggle> actorTogglesUid = new Dictionary<Atom, MyUIDynamicToggle>();
        public IEnumerable<Atom> actors => persons.Select(x => x.atom).Concat(dildos).Concat(toys);

        public static bool buttonHovered;

        public static Color navColor = new Color(0.55f, 0.90f, 1f);
        public static Color warningColor = new Color(1f, 0.60f, 0.60f);
        public static Color severeWarningColor = new Color(1f, 0.36f, 0.32f);
        
        private void OnEnable()
        {
            if(canvas != null) canvas.gameObject.SetActive(true);
            if(currentPose != null) currentPose.selected = true;
        }

        public override void Init()
        {
            try
            {
                singleton = this;
                atom = containingAtom;
                // scenename = GetSceneName();
                UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements);
                Utils.OnInitUI(CreateUIElement);
                // FileManagerSecure.CreateDirectory(managedStorePath);
                FileManagerSecure.CreateDirectory(libraryPath);
                FileManagerSecure.CreateDirectory(backUpStorePath);
                FileManagerSecure.CreateDirectory(tmpPath);
                
                sceneType = GetSceneContent(true);

                posX.setCallbackFunction += val => PositionCanvas(posX.val, posY.val);
                posY.setCallbackFunction += val => PositionCanvas(posX.val, posY.val);
                
                SetCameraRig();
                if (isVR) camMode.val = "Snap";
                // SuperController.singleton.foc
                // SuperController.singleton.FocusOnController(containingAtom.mainController);
                
                SuperController.singleton.onAtomAddedHandlers += OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers += OnAtomRemoved;
                SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;

                screenWidth = Screen.width;
                screenHeight = Screen.height;
                InitCanvas();

                ioSystem.saveDir = libraryPath;
                // ioSystem.defaultName = $"/{sceneType}/{scenename}/MyPose";
                ioSystem.store = () =>
                {
                    currentPose.UpdateJSON();
                    return currentPose.pose;
                };
                ioSystem.onSaved += val =>
                {
                    currentPose.StoreImage(val);
                };
                ioSystem.onLoaded += (path, jc) =>
                {
                    LoadPoseFromJSON(jc);
                };
                

                reapplyCurrent = new JSONStorableAction("Reapply Current Pose", () => currentPose.Apply());
                nextPose = new JSONStorableAction("Next Pose", () =>
                {
                    var pose = poses.FirstOrDefault(x => x.id > currentPose.id);
                    pose?.Apply();
                });
                previousPose = new JSONStorableAction("Previous Pose", () => {
                    var pose = poses.LastOrDefault(x => x.id < currentPose.id);
                    pose?.Apply();
                });
                randomPose = new JSONStorableAction("Random Pose", () =>
                {
                    poses.TakeRandom(currentPose.id).Apply();
                });
                nextCam = new JSONStorableAction("Next Cam", () =>
                {
                    currentPose.ApplyNextCamAngle();
                });
                previousCam = new JSONStorableAction("Previous Cam", () =>
                {
                    currentPose.ApplyPreviousCamAngle();
                });
                randomCam = new JSONStorableAction("Random Cam", () =>
                {
                    currentPose.ApplyRandomCamAngle();
                });
                printInputConfig = new JSONStorableAction("Print Input Config", PrintInputConfig);
                

                foreach (var rb in atom.rigidbodies)
                {
                    if(
                        rb.name == "hip" || 
                       rb.name == "chest" || 
                       rb.name == "head" || 
                       rb.name == "lThigh" || 
                       rb.name == "rThigh" ||
                       rb.name == "lShin" || 
                       rb.name == "rShin" ||
                       rb.name == "lShldr" ||
                       rb.name == "rShldr" ||
                       rb.name == "lHand" ||
                       rb.name == "rHand" ||
                       rb.name == "lForeArm" ||
                       rb.name == "rForeArm" ||
                       rb.name == "lFoot" ||
                       rb.name == "rFoot"
                       ) 
                        forceTargets.Add(rb);
                }

                var rBody = forceTargets.First(x => x.name == "head");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "chest");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "hip");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "lShldr");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "rShldr");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "lForeArm");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "rForeArm");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "lHand");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "rHand");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "lThigh");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "rThigh");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "lShin");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "rShin");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "lFoot");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                rBody = forceTargets.First(x => x.name == "rFoot");
                forceTargets.Remove(rBody);
                forceTargets.Add(rBody);
                
                
                GetNeighbors();
                // neighbors[atom.rigidbodies.First(x => x.name == "hip")].ToList().ForEach(x => x.name.Print());
                
                // forceTargets.ForEach(x => x.Print());
                poseChooser.RegisterNonRestore();
                camChooser.RegisterNonRestore();

                applyIdles.RegisterNonRestore();
                restoreRoot.RegisterNonRestore();
                restoreHeadRotation.RegisterNonRestore();
                restoreOtherPersons.RegisterNonRestore();
                restoreDildos.RegisterNonRestore();
                restoreToys.RegisterNonRestore();
                
                showButtons.RegisterNonRestore();
                showThumbnails.RegisterNonRestore();
                
                camMode.RegisterNonRestore();
                cinematicCamModeChooser.RegisterNonRestore();
                cinematicPoseModeChooser.RegisterNonRestore();
                cinematicEnabled.RegisterNonRestore();
                
                RegisterAction(reapplyCurrent);
                RegisterAction(nextCam);
                RegisterAction(previousCam);
                RegisterAction(randomCam);
                RegisterAction(nextPose);
                RegisterAction(previousPose);
                RegisterAction(randomPose);
                RegisterAction(printInputConfig);
                
                // ReloadFromScene();
                CreateUI();
                DeferredLoadFromScene().Start();
                
                
                // atom.rigidbodies.First(x => x.name == "lMid2").GetComponent<DAZBone>().baseJointRotation = new Vector3(0f,0f,0f);
                // cameraRig.rotation = Quaternion.identity;
                // cameraRig.position = Vector3.zero;

                // test1 = GameObject.CreatePrimitive(PrimitiveType.Sphere);
                // test2 = GameObject.CreatePrimitive(PrimitiveType.Sphere);
                // test3 = GameObject.CreatePrimitive(PrimitiveType.Sphere);
                // test1.GetComponent<Collider>().enabled = false;
                // test2.GetComponent<Collider>().enabled = false;
                // test3.GetComponent<Collider>().enabled = false;
                // test1.transform.localScale = test2.transform.localScale = test3.transform.localScale = .05f * Vector3.one;
                // test1.GetComponent<Renderer>().material.shader = test2.GetComponent<Renderer>().material.shader = test3.GetComponent<Renderer>().material.shader = FillMeUp.debugShader;
                // test1.GetComponent<Renderer>().material.color = test3.GetComponent<Renderer>().material.color = Color.magenta;
                // test2.GetComponent<Renderer>().material.color = Color.green;
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
        }

        private JSONStorableAction ToggleAction(JSONStorableBool jbool)
        {
            return new JSONStorableAction("Toggle " + jbool.name, () => jbool.val = !jbool.val);
        }
        
        public void OnBindingsListRequested(List<object> bindings)
        {
            bindings.Add(new Dictionary<string, string>
            {
                { "CheesyFX", "PoseMe" }
            });

            bindings.Add(reapplyCurrent);
            bindings.Add(nextCam);
            bindings.Add(previousCam);
            bindings.Add(randomCam);
            bindings.Add(nextPose);
            bindings.Add(previousPose);
            bindings.Add(randomPose);
            bindings.Add(printInputConfig);

            bindings.Add(ToggleAction(cinematicEnabled));
            
            bindings.Add(ToggleAction(applyIdles));
            bindings.Add(ToggleAction(restoreRoot));
            bindings.Add(ToggleAction(restoreHeadRotation));
            bindings.Add(ToggleAction(restoreOtherPersons));
            bindings.Add(ToggleAction(restoreDildos));
            bindings.Add(ToggleAction(restoreToys));
            
            bindings.Add(ToggleAction(showButtons));
            bindings.Add(ToggleAction(showThumbnails));
        }

        public void CreateUI()
        {
            tabbar = UIManager.CreateTabBar(new[] { "Pose", "General", "Cinematic", "Input" }, SelectTab);
            tabbar.SelectTab(lastTabId);

        }

        private void SelectTab(int id)
        {
            lastTabId = id;
            RemoveUIElements(UIElements);
            UIElements.Clear();
            if(id == 0) CreatePoseUI();
            else if (id == 1) CreateGeneralUI();
            else if (id == 2) CreateCinematicUI();
            else CreateInputUI();
        }

        private void CreateCinematicUI()
        {
            cinematicPoseModeChooser.CreateUI(UIElements, false, chooserType:0);
            cinematicCamModeChooser.CreateUI(UIElements, true, chooserType:0);
            camTimeMean.CreateUI(UIElements);
            camTimeDelta.CreateUI(UIElements, true);
            poseTimeMean.CreateUI(UIElements);
            poseTimeDelta.CreateUI(UIElements, true);
            camSpeedMean.CreateUI(UIElements);
            camSpeedDelta.CreateUI(UIElements, true);
            
            randomizeCamSpeed.CreateUI(UIElements, true);
            cinematicEnabled.CreateUI(UIElements);
        }

        private void CreateInputUI()
        {
            leftClickChooser.CreateUI(UIElements, rightSide: false, chooserType: 0);
            rightClickChooser.CreateUI(UIElements, rightSide: false, chooserType: 0);
            middleClickChooser.CreateUI(UIElements, rightSide: false, chooserType: 0);
            
            dragDownChooser.CreateUI(UIElements, rightSide: true, chooserType: 0);
            dragUpChooser.CreateUI(UIElements, rightSide: true, chooserType: 0);
            dragLeftChooser.CreateUI(UIElements, rightSide: true, chooserType: 0);
            dragRightChooser.CreateUI(UIElements, rightSide: true, chooserType: 0);

            printInputConfigOnFirstHover.CreateUI(UIElements);
        }

        private void CreateGeneralUI()
        {
            camMode.CreateUI(UIElements, chooserType:0);
            smoothCamSpeedJ.CreateUI(UIElements);
            smoothCamBezierStrength.CreateUI(UIElements);
            applyIdles.CreateUI(UIElements);
            manageForces.CreateUI(UIElements);
            restoreHeadRotation.CreateUI(UIElements);
            restoreRoot.CreateUI(UIElements);
            restoreOtherPersons.CreateUI(UIElements);
            restoreDildos.CreateUI(UIElements);
            restoreToys.CreateUI(UIElements);
            
            var button = this.SetupButton("Reload From Last Saved Scene", false, () => ReloadFromScene());
            button.GetComponentInChildren<Image>().color = severeWarningColor;
            UIElements.Add(button);
            button = this.SetupButton("Clear All Poses", false, ClearPoses);
            button.GetComponentInChildren<Image>().color = severeWarningColor;
            UIElements.Add(button);
            
            // this.SetupButton("Restore Defaults", restoreAllFromDefaultsAction.actionCallback.Invoke, severeWarningColor, UIElements, false);
            
            autoBackUpDeleted.CreateUI(UIElements);
            ((UIDynamicSlider)maxRows.CreateUI(UIElements, rightSide: true)).slider.wholeNumbers = true;
            buttonSizeJ.CreateUI(UIElements, rightSide:true);
            buttonTransparency.CreateUI(UIElements, rightSide: true);
            showButtons.CreateUI(UIElements, true);
            showThumbnails.CreateUI(UIElements, true);
            
        }

        public void RegisterActorToggles()
        {
            if (tabbar.id > 0 || IdleUIProvider.uiOpen || currentPose == null) return;
            foreach (var actor in actors)
            {
                actorTogglesUid[actor].RegisterBool(currentPose.actorToggles[actor]);
                actorTogglesUid[actor].SetColor(!currentPose.pose["pose"].AsObject.HasKey(actor.uid) ? warningColor : Color.white);
            }
            poseApplyIdlesUid.RegisterBool(currentPose.poseIdle.applyIdles);
        }

        private void CreatePoseUI()
        {
            UIDynamicButton button;
            poseChooser.CreateUI(UIElements, rightSide: false, chooserType: 0);
            UIElements.Add(this.SetupButton("Add Pose", false, () => AddPose()));
            this.SetupButton("Update Pose", UpdatePose, warningColor, UIElements);
            UIElements.Add(Utils.SetupTwinButton(this,
                "Move Pose Up", () => MovePose(true),
                "Move Pose Down", () => MovePose(false), false)
            );
            onPoseEnterTriggerButton = CreateButton("On Pose Enter", false);
            onPoseEnterTriggerButton.buttonColor = new Color(0.45f, 1f, 0.45f);
            onPoseExitTriggerButton = CreateButton("On Pose Exit", false);
            onPoseExitTriggerButton.buttonColor = new Color(0.45f, 1f, 0.45f);
            UIElements.Add(onPoseEnterTriggerButton);
            UIElements.Add(onPoseExitTriggerButton);
            
            button = this.SetupButton("Delete Pose", false, DeletePose);
            button.GetComponentInChildren<Image>().color = severeWarningColor;
            UIElements.Add(button);
            
            UIElements.Add(this.SetupButton("Save Pose To Library", false, () =>
            {
                if(currentPose == null) return;
                var scenename = sceneName;
                if(scenename == "") ioSystem.defaultName = $"/{sceneType}/MyScene/{currentPose.id:000}";
                else ioSystem.defaultName = $"/{sceneType}/{scenename}/{currentPose.id:000}";
                ioSystem.UISaveJSONDialog();
            }));
            button = this.SetupButton("Import Pose From Library");
            UIElements.Add(button);
            ioSystem.RegisterLoadButton(button.button);

            // if (poses.Count == 0) return;
            UIElements.Add(this.SetupButton("Configure Idles", false, () => {
                IdleUIProvider.CreateUI();
            }));
            applyIdles.CreateUI(UIElements);
            poseApplyIdlesUid = UIManager.CreateReusableUIDynamicToggle();
            UIElements.Add(poseApplyIdlesUid);
            // poseRandomizeCamUid = UIManager.CreateReusableUIDynamicToggle();
            // UIElements.Add(poseRandomizeCamUid);
            // if (currentPose != null) poseRandomizeCamUid.RegisterBool(currentPose.randomizeAngles);
            
            var textfield = CreateTextField(actorsJ);
            textfield.ForceHeight(75f);
            UIElements.Add(textfield);

            foreach (var actor in actors)
            {
                var uid = UIManager.CreateReusableUIDynamicToggle(false, actor.uid);
                actorTogglesUid[actor] = uid;
                // if(currentPose != null) actorTogglesUid[actor].RegisterBool(currentPose.actorToggles[actor]);
                UIElements.Add(uid);
            }
            RegisterActorToggles();

            
            camChooser.CreateUI(UIElements, rightSide: true, chooserType: 0);
            UIElements.Add(this.SetupButton("Add Cam", true, () => currentPose.AddCamAngle()));
            this.SetupButton("Update Cam", () => currentPose.UpdateAngle(), warningColor, UIElements, true);
            UIElements.Add(Utils.SetupTwinButton(this, 
                "Move Cam Up", () => currentPose.MoveAngle(true),
                "Move Cam Down", () => currentPose.MoveAngle(false), true)
            );
            onAngleEnterTriggerButton = CreateButton("On Cam Enter", true);
            onAngleEnterTriggerButton.buttonColor = new Color(0.45f, 1f, 0.45f);
            onAngleExitTriggerButton = CreateButton("On Cam Exit", true);
            onAngleExitTriggerButton.buttonColor = new Color(0.45f, 1f, 0.45f);
            UIElements.Add(onAngleEnterTriggerButton);
            UIElements.Add(onAngleExitTriggerButton);
            
            button = this.SetupButton("Delete Cam", true, () => currentPose.DeleteAngle());
            button.GetComponentInChildren<Image>().color = severeWarningColor;
            UIElements.Add(button);

            UIImageButton = this.SetupButton("", true);
            UIImageButton.ForceHeight(512f);
            var cb = UIImageButton.button.colors;
            cb.pressedColor = cb.highlightedColor = cb.normalColor;
            UIImageButton.button.colors = cb;
            UIImageButton.buttonText.alignment = TextAnchor.LowerRight;
            UIImageButton.buttonText.fontSize = 25;
            UIImageButton.buttonText.color = Color.white;
            UIImage = UIImageButton.GetComponentInChildren<Image>();
            UIImage.color = Color.white;
            UIElements.Add(UIImageButton);
            var clickListener = UIImageButton.gameObject.AddComponent<ClickListener>();
            // clickListener.onLeftClick.AddListener(() => currentPose.ApplyNextCamAngle());
            clickListener.onLeftClick.AddListener(onLeftClick);
            clickListener.onRightClick.AddListener(() =>
            {
                currentPose.ApplyPreviousCamAngle();
            });
            clickListener.onMiddleClick.AddListener(() => currentPose.Apply(false));
            clickListener.onPointerEnter.AddListener(() => {
                GlobalSceneOptions.singleton.disableNavigation = true;
                SuperController.singleton.disableNavigation = true;
                buttonHovered = true;
            });
            clickListener.onPointerExit.AddListener(() => {
                GlobalSceneOptions.singleton.disableNavigation = false;
                SuperController.singleton.disableNavigation = false;
                buttonHovered = false;
            });

            
            textfield = CreateTextField(
                new JSONStorableString("warning",
                    "Use <b>ONLY ONE (1)</b> instance of PoseMe per scene! All actors will be posed automatically."),
                true);
            textfield.ForceHeight(100f);
            UIElements.Add(textfield);
            if(currentPose != null) currentPose.SetUIImage();
        }

        public void ClearUI()
        {
            RemoveUIElements(leftUIElements.Select(x => (object) x.GetComponent<UIDynamic>()).ToList());
            RemoveUIElements(rightUIElements.Select(x => (object) x.GetComponent<UIDynamic>()).ToList());
            UIElements.Clear();
        }

        private IEnumerator DeferredLoadFromScene()
        {
            yield return FillMeUp.waitForPose;
            ReloadFromScene(true);
        }

        public override void InitUI()
        {
            base.InitUI();
            if(UITransform == null || uiListener != null) return;
            uiListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
            uiListener.onEnabled.AddListener(() => UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements));
            uiListener.onEnabled.AddListener(() => Utils.OnInitUI(CreateUIElement));
            uiListener.onDisabled.AddListener(() => IdleUIProvider.AxisSetActive(false));
        }

        private float camTimer;
        private float poseTimer;
        public static float currentCamSpeed;
        private readonly WaitForEndOfFrame waitForEndOfFrame = new WaitForEndOfFrame();
        
        private void Update()
        {
            if (poses.Count > 0)
            {
                if(screenWidth != Screen.width || screenHeight != Screen.height)
                {
                    PositionCanvas(posX.val, posY.val);
                    screenWidth = Screen.width;
                    screenHeight = Screen.height;
                }
                for (int i = 0; i < poses.Count; i++)
                {
                    var pose = poses[i];
                    pose.onPoseEnter.Update();
                    pose.onPoseExit.Update();
                }
                if(!cinematicEnabled.val) return;
                if (cinematicPoseMode > 0 && poses.Count > 1)// && currentPose.randomizeAngles.val)
                {
                    poseTimer -= Time.deltaTime;
                    if (poseTimer < 0)
                    {
                        poseTimer = NormalDistribution.GetValue(poseTimeMean.val, poseTimeDelta.val, 2);
                        if(cinematicPoseMode == 1) {nextPose.actionCallback.Invoke();}
                        else randomPose.actionCallback.Invoke();
                        camTimer += 5f;
                    }
                }

                if (cinematicCamMode > 0 && currentPose.camAngles.Count > 1)// && currentPose.randomizeAngles.val)
                {
                    camTimer -= Time.deltaTime;
                    if (camTimer < 0)
                    {
                        camTimer = NormalDistribution.GetValue(camTimeMean.val, camTimeDelta.val, 2);
                        currentCamSpeed = NormalDistribution.GetValue(camSpeedMean.val, camSpeedDelta.val, 2f);
                        DeferredApplyAngle(cinematicCamMode == 2).Start();
                        // currentPose.ApplyRandomCamAngle();
                    }
                }
            }
        }

        private IEnumerator DeferredApplyAngle(bool random)
        {
            yield return waitForEndOfFrame;
            if(random) currentPose.ApplyRandomCamAngle();
            else currentPose.ApplyNextCamAngle();
        }

        private void OnDisable()
        {
            if(canvas != null) canvas.SetActive(false);
            Pose.smoothCam.Stop();
            currentPose.selected = false;
        }

        private void OnDestroy()
        {
            Destroy(uiListener);
            SuperController.singleton.onAtomAddedHandlers -= OnAtomAdded;
            SuperController.singleton.onAtomRemovedHandlers -= OnAtomRemoved;
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            Destroy(canvas);
            poses.ForEach(x => x.Destroy());
            SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
        }

        private JSONArray Store()
        {
            var ja = new JSONArray();
            for (int i = 0; i < poses.Count; i++)
            {
                poses[i].UpdateJSON();
                ja.Add(poses[i].pose);
            }
            return ja;
        }
        
        private void Store(JSONClass jc)
        {
            poseChooser.Store(jc);
            camChooser.Store(jc);
            camMode.Store(jc);
            cinematicCamModeChooser.Store(jc);
            cinematicPoseModeChooser.Store(jc);
            smoothCamSpeedJ.Store(jc);
            smoothCamBezierStrength.Store(jc);
            
            camTimeMean.Store(jc);
            camTimeDelta.Store(jc);
            poseTimeMean.Store(jc);
            poseTimeDelta.Store(jc);
            camSpeedMean.Store(jc);
            camSpeedDelta.Store(jc);

            showThumbnails.Store(jc);
            showButtons.Store(jc);
            maxRows.Store(jc);
            buttonSizeJ.Store(jc);
            buttonTransparency.Store(jc);
            
            printInputConfigOnFirstHover.Store(jc);
            applyIdles.Store(jc);
            restoreRoot.Store(jc);
            restoreHeadRotation.Store(jc);
            restoreOtherPersons.Store(jc);
            restoreDildos.Store(jc);
            restoreToys.Store(jc);
            manageForces.Store(jc);

            leftClickChooser.Store(jc);
            rightClickChooser.Store(jc);
            middleClickChooser.Store(jc);
            dragDownChooser.Store(jc);
            dragLeftChooser.Store(jc);
            dragUpChooser.Store(jc);
            dragRightChooser.Store(jc);
        }
        
        private void Load(JSONClass jc)
        {
            poseChooser.Load(jc);
            camChooser.Load(jc);
            camMode.Load(jc);
            cinematicCamModeChooser.Load(jc);
            cinematicPoseModeChooser.Load(jc);
            smoothCamSpeedJ.Load(jc);
            smoothCamBezierStrength.Load(jc);
            
            camTimeMean.Load(jc);
            camTimeDelta.Load(jc);
            poseTimeMean.Load(jc);
            poseTimeDelta.Load(jc);
            camSpeedMean.Load(jc);
            camSpeedDelta.Load(jc);

            showThumbnails.Load(jc);
            showButtons.Load(jc);
            maxRows.Load(jc);
            buttonSizeJ.Load(jc);
            buttonTransparency.Load(jc);
            
            printInputConfigOnFirstHover.Load(jc);
            applyIdles.Load(jc);
            restoreRoot.Load(jc);
            restoreHeadRotation.Load(jc);
            restoreOtherPersons.Load(jc);
            restoreDildos.Load(jc);
            restoreToys.Load(jc);
            manageForces.Load(jc);

            leftClickChooser.Load(jc);
            rightClickChooser.Load(jc);
            middleClickChooser.Load(jc);
            dragDownChooser.Load(jc);
            dragLeftChooser.Load(jc);
            dragUpChooser.Load(jc);
            dragRightChooser.Load(jc);
        }
        
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, true);
            Store(jc);
            jc["poses"] = Store();
            return jc;
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true,
            bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (!physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                needsStore = true;
                if (!jc.HasKey("poses")) return;
                ClearPoses();
                var ja = jc["poses"].AsArray;
                foreach (JSONClass poseJc in ja.Childs)
                {
                    LoadPoseFromJSON(poseJc, false);
                }
                poses.Last().Apply();
                Load(jc);
                OnShowThumbnailChanged(showThumbnails.val);
                SyncAngleChooser();
            }
            if(!initialized) SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);
            initialized = true;
        }

        private void ReloadFromScene(bool init = false)
        {
            var sceneName = SuperController.singleton.LoadedSceneName;
            if (sceneName == null)
            {
                if(!init) "Scene not loaded from a json.".Print();
                return;
            }

            var sceneJson = SuperController.singleton.LoadJSON(sceneName)["atoms"].Childs
                .FirstOrDefault(x => x["id"].Value == containingAtom.uid);
            if (sceneJson == null)
            {
                if(!init) $"Atom {containingAtom.uid} not found in the loaded scene json. Did you rename the person?".Print();
                return;
            }
            sceneJson = sceneJson["storables"].Childs.
                FirstOrDefault(x => x["id"].Value.EndsWith("CheesyFX.PoseMe"));
            if (sceneJson == null)
            {
                if(!init) $"Atom {containingAtom.uid} doesn't have PoseMe stored in the loaded scene json.".Print();
                return;
            }
            // SuperController.singleton.LoadedSceneName.Print();
            var pluginJSON = sceneJson.AsObject;
            if (!pluginJSON.HasKey("poses") || !pluginJSON["poses"].Childs.Any())
            {
                if(!init) $"Scene {sceneName} doesn't have poses.".Print();
                return;
            }
            LateRestoreFromJSON(pluginJSON, true, true, true);
        }

        public void OnAtomRename(string oldUid, string newUid)
        {
            for (int i = 0; i < poses.Count; i++)
            {
                var pose = poses[i];
                pose.SyncAtomNames(oldUid, newUid);
                pose.SyncActorToggleNames();
            }
            RegisterActorToggles();
            actorsJ.val = currentPose.actors;
        }

        private void OnAtomAdded(Atom atom)
        {
            try
            {
                if (atom.type != "Person" && !atom.IsToyOrDildo()) return;
                sceneType = GetSceneContent();
                ioSystem.defaultName = $"/{sceneType}/MyPose";
                for (int i = 0; i < poses.Count; i++)
                {
                    poses[i].actorToggles[atom] = new JSONStorableBool(atom.uid, true);
                }
                
                if (tabbar.id == 0 && !IdleUIProvider.uiOpen)
                {
                    actorTogglesUid[atom] = UIManager.CreateReusableUIDynamicToggle();
                    UIElements.Add(actorTogglesUid[atom]);
                    RegisterActorToggles();
                }
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
                throw;
            }
        }
        
        private void OnAtomRemoved(Atom atom)
        {
            try
            {
                if (atom.type != "Person" && !atom.IsToyOrDildo()) return;
                sceneType = GetSceneContent();
                ioSystem.defaultName = $"/{sceneType}/MyPose";
                for (int i = 0; i < poses.Count; i++)
                {
                    poses[i].actorToggles.Remove(atom);
                }

                if (tabbar.id == 0 && !IdleUIProvider.uiOpen)
                {
                    RemoveUIElement(actorTogglesUid[atom]);
                    UIElements.Remove(actorTogglesUid[atom]);
                }
                actorTogglesUid.Remove(atom);
                RegisterActorToggles();
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
        }

        public static UnityAction onLeftClick = () => currentPose.ApplyNextCamAngle();
        public static UnityAction onRightClick = () => currentPose.ApplyPreviousCamAngle();
        public static UnityAction onMiddleClick = () => currentPose.Apply();
        public static UnityAction onDragUp = () => currentPose.ApplyPreviousCamAngle();
        public static UnityAction onDragDown = () => currentPose.ApplyNextCamAngle();
        public static UnityAction onDragLeft = () => currentPose.ApplyRandomCamAngle();
        public static UnityAction onDragRight = () => currentPose.ApplyRandomCamAngle();

        private static void SyncInput(string val, ref UnityAction action)
        {
            switch (val)
            {
                case "None":
                {
                    action = delegate {};
                    break;
                }
                case "Next Cam":
                {
                    action = () => currentPose.ApplyNextCamAngle();
                    break;
                }
                case "Previous Cam":
                {
                    action = () => currentPose.ApplyPreviousCamAngle();
                    break;
                }
                case "Random Cam":
                {
                    action = () => currentPose.ApplyRandomCamAngle();
                    break;
                }
                case "Apply Pose":
                {
                    action = () => currentPose.Apply();
                    break;
                }
            }
        }

        public static void SyncAngleChooser()
        {
            if(currentPose == null) return;
            camChooser.choices = currentPose.camAngles.Select(x => x.id.ToString()).ToList();
        }

        public static List<Atom> toys = new List<Atom>();
        private static string GetSceneContent(bool init = false)
        {
            int femCount = 0;
            int futaCount = 0;
            int maleCount = 0;
            dildos.Clear();
            toys.Clear();
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                if (atom.type == "Person")
                {
                    var person = persons.First(x => x.uid == atom.uid);
                    if (person.genderListener.dcs.gender == DAZCharacterSelector.Gender.Female)
                    {
                        if (person.genderListener.isFuta) futaCount++;
                        else femCount++;
                    }
                    else maleCount++;
                }
                else if (atom.type == "Dildo")
                {
                    dildos.Add(atom);
                }
                else if (atom.IsToy())
                {
                    toys.Add(atom);
                }
            }
            var newsceneType = new string('F', femCount)+new string('M', maleCount)+new string('D', dildos.Count)+new string('T', toys.Count);
            if (!init && newsceneType != sceneType && poses.Count > 0)
            {
                "PoeseMe: Actors have changed. Please update your poses.".Print();
            }
            return newsceneType;
        }

        private void LoadPoseFromJSON(JSONClass jc, bool apply = true)
        {
            var pose = AddPose(jc);
            if(apply) pose.Apply();
        }

        private Pose AddPose(JSONClass jc = null)
        {
            int id;
            if (poses.Count == 0) id = 0;
            else id = poses.Select(x => x.id).Max()+1;
            Pose pose;
            if (jc != null) pose = new Pose(id, jc);
            else pose = new Pose(id);
            poses.Add(pose);
            var choices = poseChooser.choices;
            choices.Add(id.ToString());
            poseChooser.choices = null;
            poseChooser.choices = choices;
            poseChooser.valNoCallback = poseChooser.choices.Last();
            LayoutButtons();
            RegisterActorToggles();
            return pose;
        }

        private static void MovePose(bool up)
        {
            if(currentPose == null) return;
            int i = currentPose.id;
            if (up)
            {
                if (i == 0) return;
                var upperPose = poses.LastOrDefault(x => x.id < i);
                if (upperPose != null)
                {
                    poses[i-1] = currentPose;
                    poses[i] = upperPose;
                    upperPose.id++;
                    currentPose.id--;
                }
                // poseChooser.valNoCallback = (i - 1).ToString();
            }
            else
            {
                if(i == poses.Count - 1) return;
                var lowerPose = poses.FirstOrDefault(x => x.id > i);
                if (lowerPose != null)
                {
                    poses[i+1] = currentPose;
                    poses[i] = lowerPose;
                    lowerPose.id--;
                    currentPose.id++;
                }
            }
            RenamePoses();
            LayoutButtons();
            poseChooser.valNoCallback = currentPose.id.ToString();
        }

        private void ClearPoses()
        {
            if (poses.Count == 0) return;
            foreach (var pose in poses)
            {
                pose.Destroy();
            }
            poseChooser.valNoCallback = "";
            poseChooser.choices = new List<string>();
            currentPose = null;
            poses.Clear();
        }

        private static void DeletePose()
        {
            if (currentPose == null) return;
            currentPose.Destroy();
            poses.Remove(currentPose);
            RenamePoses();
            poseChooser.choices = poses.Select(x => x.id.ToString()).ToList();
            poseChooser.valNoCallback = "";
            if(autoBackUpDeleted.val) currentPose.StoreBackup();
            LayoutButtons();
            UIImage.sprite = null;
            currentPose = null;
        }

        private static void RenamePoses()
        {
            for (int i = 0; i < poses.Count; i++)
            {
                poses[i].id = i;
            }
        }
        
        private static void UpdatePose()
        {
            var pose = poses[int.Parse(poseChooser.val)];
            pose.UpdatePose();
        }

        private static void ApplyPose(string id)
        {
            int i;
            if(int.TryParse(id, out i) && i < poses.Count){
                poses[i].Apply();
            }
        }

        private static void ApplyCam(string id)
        {
            if (currentPose == null) return;
            int i;
            if (int.TryParse(id, out i) && i < currentPose.camAngles.Count)
            {
                currentPose.camAngles[i].Apply();
            }
        }

        private static void OnShowThumbnailChanged(bool val)
        {
            for (int i = 0; i < poses.Count; i++)
            {
                poses[i].SetButtonImage();
            }

            SyncButtonSize(buttonSizeJ.val);
        }

        private static void SyncButtonSize(float val)
        {
            buttonWidth = buttonHeight = showThumbnails.val? val : val * .75f;
            LayoutButtons();
            if(val < buttonSizeJ.defaultVal)
            {
                for (int i = 0; i < poses.Count; i++)
                {
                    poses[i].SetButtonText();
                }
            }
        }

        private static string sceneName
        {
            get
            {
                var loadedScenePath = SuperController.singleton.LoadedSceneName;
                if (loadedScenePath == null) return "";
                var idx = loadedScenePath.LastIndexOf("/", StringComparison.Ordinal);
                if (idx == -1) return "";
                return loadedScenePath.Substring(idx + 1).Replace(".json", "");
            }
        }
        
        public static bool inputConfigPrinted;
        public static void PrintInputConfig()
        {
            if (inputConfigPrinted || !printInputConfigOnFirstHover.val) return;
            inputConfigPrinted = true;
            string info = "PoseMe: Input config for the buttons\n\n"+
                          $"Left click: {leftClickChooser.val}\n" +
                          $"Right click: {rightClickChooser.val}\n" +
                          $"Middle click: {rightClickChooser.val}\n" +
                          $"Drag left: {dragLeftChooser.val}\n" +
                          $"Drag right: {dragRightChooser.val}\n" +
                          $"Drag up: {dragUpChooser.val}\n" +
                          $"Drag down: {dragDownChooser.val}\n\n"+
                          $"Go to '{atom.name}/BodyLanguage/PoseMe/Input' to review or change.";
            info.Print();
        }
        
        private void SetCameraRig()
        {
            if (SuperController.singleton.isOpenVR)
            {
                camera = SuperController.singleton.ViveCenterCamera.transform;
                isVR = true;
            }
            else if (SuperController.singleton.isOVR)
            {
                camera = SuperController.singleton.OVRCenterCamera.transform;
                isVR = true;
            }
            else
            {
                camera = SuperController.singleton.MonitorCenterCamera.transform;
            }

            cameraRig = camera.parent.parent;
            // cameraRig.Draw();
            // // (SuperController.singleton.navigationRig == cameraRig).Print();
            // cameraRig.transform.position.Print();
            // camera.transform.position.Print();
        }

        private static string GetScenePackageUid()
        {
            var loadedScenePath = SuperController.singleton.LoadedSceneName;
            if (loadedScenePath == null) return "";
            var idx = loadedScenePath.IndexOf(":/", StringComparison.Ordinal);
            if (idx == -1) return "";
            return loadedScenePath.Substring(0, idx+2);
        }
        
        public static IEnumerator CreateAtomCo(string type, string basename, Action<Atom> onAtomCreated)
        {
            string uid = NewUID(basename);
            yield return SuperController.singleton.AddAtomByType(type, uid, true);
            var atom = SuperController.singleton.GetAtomByUid(uid);
            if(atom == null)
            {
                throw new NullReferenceException("atom was not created");
            }
            SuperController.singleton.RenameAtom(atom, basename);
            onAtomCreated(atom);
        }

        public static string NewUID(string basename)
        {
            var uids = new HashSet<string>(SuperController.singleton.GetAtomUIDs());
            if(!uids.Contains(basename))
            {
                return basename;
            }

            for(int i = 2; i < 1000; i++)
            {
                string uid = $"{basename}#{i}";
                if(!uids.Contains(uid))
                {
                    return uid;
                }
            }

            // you don't really want to be here 
            return basename + Guid.NewGuid();
        }
        
        public void ForceSaveJSON(JSONClass jc, string path)
        {
            var confirmButtonTransform = SuperController.singleton.transform
                .Find("WorldScaleAdjust/HUDForLogs/UserConfirmCanvas/UserConfirmPluginActionPanel(Clone)/Panel/ConfirmButton");
            if(confirmButtonTransform == null) return;
            confirmButtonTransform.GetComponent<Button>().onClick.Invoke();
        }

        // StartCoroutine(CreateAtomCo("Empty", "shiny", atom => {
        //     // atom reference available here
        // }));

        private void InitCanvas()
        {
            Transform MainWindowUI = SuperController.singleton.MonitorModeAuxUI.gameObject.transform.parent;
            if (MainWindowUI == null)
            {
                SuperController.LogError("MainWindowUI not found.");
                return;
            }

            Transform panel = MainWindowUI.GetChild(0);
            if (panel == null)
            {
                SuperController.LogError("Panel not found.");
                return;
            }

            canvas = new GameObject("BL_ButtonCanvas");
            canvas.transform.SetParent(panel, false);
            var rt = canvas.AddComponent<RectTransform>();
            rt.anchorMax = new Vector2(0,0);
            rt.anchorMin = new Vector2(0, 0);
            rt.offsetMax = new Vector2(0, 0);
            rt.offsetMin = new Vector2(0, 0);
            canvas.gameObject.layer = panel.gameObject.layer;
            PositionCanvas(1, 1);
        }

        private void GetNextShot()
        {
            Pose pose;
            CamAngle angle;
            if (currentPose.currentAngle.id < currentPose.camAngles.Count - 1)
            {
                pose = currentPose;
                angle = currentPose.camAngles[currentPose.currentAngle.id + 1];
            }
            else if (currentPose.id < poses.Count - 1)
            {
                pose = poses[currentPose.id + 1];
                angle = pose.camAngles[0];
            }
        }

        public static void LayoutButtons()
        {
            maxRows.max = (int)(2f * Screen.height / (buttonHeight+buttonSpacing));
            int columnCount = (int)Math.Ceiling((float)poses.Count / maxRows.val);
            int row, column;
            for (var i = 0; i < poses.Count; i++)
            {
                var rt = poses[i].buttonRT;
                column = i % columnCount;
                row = i / columnCount;
                rt.offsetMax = new Vector2(-(columnCount-column-1)*(buttonWidth+buttonSpacing), -row * (buttonHeight+buttonSpacing));
                rt.offsetMin = rt.offsetMax - new Vector2(buttonWidth, buttonHeight);
            }
        }
        
        private void PositionCanvas(float width, float height)
        {
            RectTransform rt = canvas.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(2f*width*Screen.width-buttonSpacing, 2f*height*Screen.height-buttonSpacing);
            LayoutButtons();
        }
        
        public Dictionary<Rigidbody, IEnumerable<FreeControllerV3>> neighbors = new Dictionary<Rigidbody, IEnumerable<FreeControllerV3>>();
        public Dictionary<Rigidbody, IEnumerable<FreeControllerV3>> children = new Dictionary<Rigidbody, IEnumerable<FreeControllerV3>>();
        private void GetNeighbors()
        {
            foreach (var target in forceTargets)
            {
                switch (target.name)
                {
                    case "hip":
                    {
                        neighbors[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("Thigh")||
                            x.name.Contains("abdomen")||
                            x.name.Contains("pelvis"));
                        break;
                    }
                    case "lThigh":
                    {
                        neighbors[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("rThigh")||
                            x.name.Contains("abdomen")||
                            x.name.Contains("pelvis")||
                            x.name.Contains("hip"));
                        children[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("lKnee"));
                        break;
                    }
                    case "rThigh":
                    {
                        neighbors[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("lThigh")||
                            x.name.Contains("abdomen")||
                            x.name.Contains("pelvis")||
                            x.name.Contains("hip"));
                        children[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("rKnee"));
                        break;
                    }
                    case "lShin":
                    {
                        children[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("lFoot"));
                        break;
                    }
                    case "rShin":
                    {
                        children[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("rFoot"));
                        break;
                    }
                    case "chest":
                    {
                        neighbors[target] = children[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("Arm"));
                        break;
                    }
                    case "lForeArm":
                    {
                        children[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("lHand"));
                        break;
                    }
                    case "rForeArm":
                    {
                        children[target] = containingAtom.freeControllers.Where(x =>
                            x.name.Contains("rHand"));
                        break;
                    }
                }
            }
        }

        public void RemoveUIElements(List<object> UIElements)
		{
			for (int i=0; i<UIElements.Count; ++i)
            {
                RemoveUIElement(UIElements[i]);
            }
            UIElements.Clear();
		}
        
        public void RemoveUIElement(object element)
		{
            if(element == null) return;
			if (element is JSONStorableParam)
			{
				JSONStorableParam jsp = element as JSONStorableParam;
				if (jsp is JSONStorableFloat)
					RemoveSlider(jsp as JSONStorableFloat);
				else if (jsp is JSONStorableBool)
					RemoveToggle(jsp as JSONStorableBool);
				else if (jsp is JSONStorableColor)
					RemoveColorPicker(jsp as JSONStorableColor);
				else if (jsp is JSONStorableString)
					RemoveTextField(jsp as JSONStorableString);
				else if (jsp is JSONStorableStringChooser)
				{
					// Workaround for VaM not cleaning its panels properly.
					JSONStorableStringChooser jssc = jsp as JSONStorableStringChooser;
					RectTransform popupPanel = jssc.popup?.popupPanel;
					RemovePopup(jssc);
					if (popupPanel != null)
						UnityEngine.Object.Destroy(popupPanel.gameObject);
				}
			}
			else if (element is UIDynamic)
			{
				UIDynamic uid = element as UIDynamic;
				if (uid is UIDynamicButton)
					RemoveButton(uid as UIDynamicButton);
                else if (uid is UIDynamicSlider)
					RemoveSlider(uid as UIDynamicSlider);
                else if (uid is UIDynamicColorPicker)
					RemoveColorPicker(uid as UIDynamicColorPicker);
				else if (uid is UIDynamicTextField)
					RemoveTextField(uid as UIDynamicTextField);
				else if (uid is UIDynamicPopup)
				{
					// Workaround for VaM not cleaning its panels properly.
					UIDynamicPopup uidp = uid as UIDynamicPopup;
					RectTransform popupPanel = uidp.popup?.popupPanel;
					RemovePopup(uidp);
					if (popupPanel != null)
						UnityEngine.Object.Destroy(popupPanel.gameObject);
				}
                else if (uid is MyUIDynamicToggle)
                {
                    var uidToggle = uid as MyUIDynamicToggle;
                    leftUIElements.Remove(uidToggle.transform);
                    rightUIElements.Remove(uidToggle.transform);
                    Destroy(uidToggle.gameObject);
                }
                else if (uid is UIDynamicToggle)
                    RemoveToggle(uid as UIDynamicToggle);
                else if (uid is UIDynamicV3Slider)
                {
                    var v3Slider = uid as UIDynamicV3Slider;
                    leftUIElements.Remove(v3Slider.transform);
                    rightUIElements.Remove(v3Slider.spacer.transform);
                    Destroy(v3Slider.spacer.gameObject);
                    Destroy(v3Slider.gameObject);
                }
                else if (uid is UIDynamicTabBar)
                {
	                var tabbar = uid as UIDynamicTabBar;
                    leftUIElements.Remove(tabbar.transform);
                    rightUIElements.Remove(tabbar.spacer.transform);
                    Destroy(tabbar.spacer.gameObject);
                    Destroy(tabbar.gameObject);
                }
                else
					RemoveSpacer(uid);
			}
        }
    }
}